import os
import json
from pwact.utils.json_operation import get_parameter, get_required_parameter
from pwact.active_learning.user_input.train_param.nn_feature_type import Descriptor
from pwact.active_learning.user_input.train_param.model_param import ModelParam
from pwact.active_learning.user_input.train_param.optimizer_param import OptimizerParam
from pwact.active_learning.user_input.train_param.work_file_param import WorkFileStructure

from pwact.active_learning.user_input.train_param.nep_param import NepParam
from pwact.utils.constant import get_atomic_name_from_str
'''
description: 
    covert the input json file to class
return {*}
author: wuxingxing
'''
class InputParam(object):
    '''
    description: 
        covert the input json file to class
    param {*} self
    param {dict} json_input: the input json file from user
    return {*}
    author: wuxingxing
    '''
    def __init__(self, json_input:dict, cmd) -> None:
        self.nep_param = None

        self.cmd = cmd
        self.inference = True if self.cmd == "test".upper() else False
        self.model_type = get_required_parameter("model_type", json_input).upper()
        # self.atom_type = get_required_parameter("atom_type", json_input)
        self.atom_type = get_atomic_name_from_str(get_required_parameter("atom_type", json_input))
        self.model_num = get_parameter("model_num", json_input, 1)
        self.recover_train = get_parameter("recover_train", json_input, True)
        self.max_neigh_num = get_parameter("max_neigh_num", json_input, 100)
        self.profiling = get_parameter("profiling", json_input, False)#not realized

        self.set_feature_params(json_input)
        self.set_workdir_structures(json_input)
        if self.inference and self.model_type in ["NN", "LINEAR"]:
            self.file_paths.nn_work = os.path.join(self.file_paths.json_dir, "work_test_dir")

        if self.model_type in ["DP", "NN", "NEP", "LINEAR", "CHEBY"]:
            self.set_model_init_params(json_input)
            self.set_default_multi_gpu_info(json_input)
            # set optimizer
            self.set_optimizer(json_input)
            if self.model_type in ["NN", "LINEAR"]:
                self.optimizer_param.batch_size = 1
                print("Warning! NN and Linear models only support single batch, automatically adjust batch_size=1.")
        # elif self.model_type in ["NEP"]:
        #     self.set_nep_in_params(json_input)
            
        elif self.model_type in ["GNN"]:
            raise Exception("GNN is not realized yet!")
        else:
            raise Exception("ERROR! {} model type not supported!".format(self.model_type))

        # DP, NN, and NEP common file paths
        self.file_paths.set_model_file_paths(json_input)
        self.file_paths.set_train_valid_file(json_input)

    '''
    description: 
        extract params from nep.in file or 'model' dict, it both exsits, prioritize taking from nep.in file
        对于NEP 参数，存在nep.in文件，则从nep.in提取，否则从json字典提取，先提取未nepparam对象，然后从中取出值，复制给模型和优化器
        需要补充 从 nep.txt 中取参数，支持对nep.txt 继续训练
    param {*} self
    param {dict} json_input
    return {*}
    author: wuxingxing
    '''    
    def set_nep_params(self, json_input:dict):
        self.model_param = ModelParam()
        nep_dict = get_parameter("model", json_input, {})
        self.model_param.set_type_embedding_net(
                                    network_size=self.descriptor.type_network_size, 
                                    bias=self.descriptor.type_bias, 
                                    resnet_dt=self.descriptor.type_resnet_dt, 
                                    activation=self.descriptor.type_activation,
                                    physical_property=self.descriptor.type_physical_property)
        self.model_param.set_nn_fitting_net(get_parameter("fitting_net",nep_dict, {}))
        self.is_dfeat_sparse = get_parameter("is_dfeat_sparse", json_input, False)  #for nn, 'true' not realized

        # nep_in_file = get_parameter("nep_in_file", json_input, None)
        nep_param = NepParam()
        # if nep_in_file is not None:
        #     nep_param.set_nep_param_from_nep_in(nep_in_file, self.atom_type)
        nep_txt_file = get_parameter("nep_txt_file", json_input, None)
        if nep_txt_file is not None:
            nep_param.set_nep_nn_c_param_from_nep_txt(nep_txt_file)
        else:
            nep_param.set_nep_param_from_json(json_input, self.atom_type)
        self.nep_param = nep_param
        self.model_param.fitting_net.network_size = nep_param.neuron
        self.descriptor.cutoff = nep_param.cutoff
        self.descriptor.n_max = nep_param.n_max
        self.descriptor.basis_size = nep_param.basis_size
        self.descriptor.l_max = nep_param.l_max
        self.descriptor.zbl = nep_param.zbl
        # self.descriptor.type_weight = nep_param.type_weight


    '''
    description: 
        params for dp and nn model training/inference work
    param {*} self
    param {dict} json_input
    return {*}
    author: wuxingxing
    '''
    def set_model_init_params(self, json_input:dict):
        # if feature_type specified at first layer of JSON, use this value as feature type in descriptor
        # nn feature types default is [3,4]
        self.feature_type = get_parameter("feature_type", json_input, None) 

        self.type_embedding = get_parameter("type_embedding",json_input, False)
        model_dict = get_parameter("model", json_input, {})
        self.descriptor = Descriptor(get_parameter("descriptor", model_dict, {}), self.model_type, self.cmd, self.feature_type, self.type_embedding)
        
        self.atom_type_dict = []
        for idx in self.atom_type:
            #the last 4 param used in function gen_config_inputfile()
            single_dict = {'type': idx, 'Rc': self.descriptor.Rmax, 'Rm': self.descriptor.Rmin, 'iflag_grid': 3, 'fact_base': 0.2, 'dR1': 0.5, "b_init":166.0}    
            self.atom_type_dict.append(single_dict)

        self.type_embedding = self.descriptor.type_embedding
        if self.model_type == "DP":
            self.model_param = ModelParam()
            self.model_param.set_type_embedding_net(
                                                    network_size=self.descriptor.type_network_size, 
                                                    bias=self.descriptor.type_bias, 
                                                    resnet_dt=self.descriptor.type_resnet_dt, 
                                                    activation=self.descriptor.type_activation,
                                                    physical_property=self.descriptor.type_physical_property)
            self.model_param.set_embedding_net(network_size=self.descriptor.network_size, 
                                               bias=self.descriptor.bias, 
                                               resnet_dt=self.descriptor.resnet_dt, 
                                               activation=self.descriptor.activation)
            self.model_param.set_dp_fitting_net(get_parameter("fitting_net",model_dict, {}))
        elif self.model_type == "NN":
            self.model_param = ModelParam()
            self.model_param.set_type_embedding_net(
                                        network_size=self.descriptor.type_network_size, 
                                        bias=self.descriptor.type_bias, 
                                        resnet_dt=self.descriptor.type_resnet_dt, 
                                        activation=self.descriptor.type_activation,
                                        physical_property=self.descriptor.type_physical_property)
            self.model_param.set_nn_fitting_net(get_parameter("fitting_net",model_dict, {}))
            self.is_dfeat_sparse = get_parameter("is_dfeat_sparse", json_input, False)  #'true' not realized
        elif self.model_type == "Linear".upper():
            pass
        elif self.model_type == "NEP".upper():
            self.set_nep_params(json_input) #  nep.in 输入的适配可能并不需要
            self.file_paths.set_nep_native_file_paths()#  nep.in 输入的适配可能并不需要
        elif self.model_type == "CHEBY".upper():
            self.model_param = ModelParam()
            self.model_param.set_nn_fitting_net(get_parameter("fitting_net",model_dict, {}))
        else: # linear
            raise Exception("model_type {} not realized yet".format(self.model_type))
    
    def set_optimizer(self, json_input:dict):
        self.optimizer_param = OptimizerParam()
        self.optimizer_param.set_optimizer( json_input ,self.nep_param)
        #

    '''
    description: 
        Reserved param interface for multi-GPU, multi-node parallel training of DP/NN models
    param {*} self
    param {dict} json_input
    return {*}
    author: wuxingxing
    '''    
    def set_default_multi_gpu_info(self, json_input:dict):
        # multi GPU train params  these params not used
        # "number of data loading workers (default: 4)
        self.workers = get_parameter("workers", json_input, 1)
        # dist training by horovod, when multi GPU train, need True,
        self.hvd = get_parameter("hvd", json_input, False)
        self.world_size = get_parameter("world_size", json_input, -1)
        self.rank = get_parameter("rank", json_input, -1)
        self.dist_url = get_parameter("dist_url", json_input, "tcp://localhost:23456")  
        self.dist_backend = get_parameter("dist_backend", json_input, "nccl")
        self.distributed = get_parameter("distributed", json_input, False)
        self.gpu = get_parameter("gpu", json_input, None)

    '''
    description: 
        set feature common params (dp/nn/nep/gnn)
    param {*} self
    param {dict} json_input
    return {*}
    author: wuxingxing
    '''    
    def set_feature_params(self, json_input:dict):
        # set feature related params
        self.valid_shuffle = get_parameter("valid_shuffle", json_input, False)
        self.data_shuffle = get_parameter("data_shuffle", json_input, True)
        self.seed = get_parameter("seed", json_input, 2023)
        self.precision = get_parameter("precision", json_input, "float64")

    '''
    description: 
    set common workdir
    param {*} self
    param {dict} json_input
    return {*}
    author: wuxingxing
    '''
    def set_workdir_structures(self, json_input:dict):
        # set file structures
        self.file_paths = WorkFileStructure(json_dir=os.getcwd(), 
                            reserve_work_dir=get_parameter("reserve_work_dir", json_input, False), 
                            reserve_feature = get_parameter("reserve_feature", json_input, False), 
                            model_type=self.model_type)

    '''
    description: 
        set dp and NN inference params
    param {*} self
    param {dict} json_input
            is_nep_txt for nep.txt file load
    return {*}
    author: wuxingxing
    '''    
    def set_test_relative_params(self, json_input:dict, is_nep_txt:bool=False):
        self.inference = True
        self.recover_train = True
        self.optimizer_param.batch_size = 1     # set batch size to 1, so that each image inference info will be saved
        self.data_shuffle = False
        self.valid_shuffle = False
        self.file_paths.set_inference_paths(json_input,is_nep_txt = is_nep_txt)
    
    '''
    description: 
        this dict is used for dp model initialization
    param {*} self
    return {*}
    author: wuxingxing
    '''
    def get_dp_net_dict(self):
        net_dict = {}
        net_dict["model_type"] = self.model_type
        net_dict["net_cfg"] = {}
        net_dict["net_cfg"][self.model_param.embedding_net.net_type] = self.model_param.embedding_net.to_dict()
        net_dict["net_cfg"][self.model_param.fitting_net.net_type] = self.model_param.fitting_net.to_dict()
        net_dict["net_cfg"][self.model_param.type_embedding_net.net_type] = self.model_param.type_embedding_net.to_dict()
        net_dict["M2"] = self.descriptor.M2
        net_dict["maxNeighborNum"] = self.max_neigh_num
        net_dict["atomType"]=self.atom_type_dict
        net_dict["training_type"] = self.precision
        net_dict["Rc_M"] = self.descriptor.Rmax
        return net_dict
    
    '''
    description: 
        this dict is used for feature generation of NN models
    param {*} self
    return {*}
    author: wuxingxing
    '''    
    def get_data_file_dict(self):
        data_file_dict = {}
        # data_file_dict = self.file_paths.get_data_file_structure()
        data_file_dict["M2"] = self.descriptor.M2
        data_file_dict["maxNeighborNum"] = self.max_neigh_num
        data_file_dict["atomType"]=self.atom_type_dict
        data_file_dict["Rc_M"] = self.descriptor.Rmax
        data_file_dict["E_tolerance"] = self.descriptor.E_tolerance
        data_file_dict["train_egroup"] = self.optimizer_param.train_egroup

        return data_file_dict
    
    def to_dict(self):
        params_dict = {}
        params_dict["model_type"] = self.model_type
        params_dict["atom_type"] = self.atom_type
        if self.model_type !="NEP":
            params_dict["max_neigh_num"] = self.max_neigh_num
        if self.seed is not None:
            params_dict["seed"] = self.seed
        if self.model_num > 1 :
            params_dict["model_num"] = self.model_num
        # params_dict["E_tolerance"] = self.descriptor.E_tolerance
        # params_dict["Rmax"] = self.descriptor.Rmax
        # params_dict["Rmin"] = self.Rmin
        # params_dict["M2"] = self.descriptor.M2
        # params_dict["data_shuffle"] = self.data_shuffle
        # if self.cmd == "train".upper():
        #     params_dict["train_valid_ratio"] = self.train_valid_ratio
        
        # params_dict["precision"] = self.precision

        # params_dict["profiling"] = self.profiling
        # params_dict["workers"] = self.workers
        # params_dict["hvd"] = self.hvd
        # params_dict["world_size"] = self.world_size
        # params_dict["rank"] = self.rank
        # params_dict["dist_url"] = self.dist_url
        # params_dict["dist_backend"] = self.dist_backend
        # params_dict["distributed"] = self.distributed

        # params_dict["inference"] = self.inference
        
        if self.cmd == "train".upper() and self.model_type != "Linear".upper():
            params_dict["recover_train"] = self.recover_train
        
        params_dict["model"] = {}
        if self.model_type in ["DP", "NN", "LINEAR", "CHEBY", "NEP"]:
            params_dict["model"]["descriptor"] = self.descriptor.to_dict()

            if self.model_type == "Linear".upper():
                params_dict["optimizer"] = self.optimizer_param.to_linear_dict()
            else:
                params_dict["model"]["fitting_net"] = self.model_param.fitting_net.to_dict_std()
                params_dict["optimizer"] = self.optimizer_param.to_dict()
        # elif self.model_type in ["NEP"]:
        #     nep_in_content = self.model_param.nep_param.to_txt()
        elif self.model_type in ["GNN"]:
            raise Exception("The model GNN not realized yet!")

        file_dir_dict = self.file_paths.to_dict()
        for key in file_dir_dict:
            params_dict[key] = file_dir_dict[key]
        return params_dict
    
    def print_input_params(self, json_file_save_name="json_all.json"):
        params_dict = self.to_dict()
        json.dump(params_dict, open(os.path.join(self.file_paths.json_dir, json_file_save_name), "w"), indent=4)
        print(params_dict)
        
def help_info():
    # 使用双线边框和加粗标题
    print("\n\033[1;36m╔" + "=" * 48 + "╗\033[0m")  # 双线上边框
    print("\033[1;36m║" + " " * 10 + "\033[1;35m PWMLFF Basic Information \033[0m" + " " * 12 + "\033[1;36m║\033[0m")  # 标题
    print("\033[1;36m╚" + "=" * 48 + "╝\033[0m")  # 双线下边框
    print(f"\033[1;32mVersion:\033[0m 2025.02")
    print(f"\033[1;32mCompatible pwdata:\033[0m >= 0.4.8")
    print(f"\033[1;32mCompatible pwact:\033[0m >= 0.2.1")
    print(f"\033[1;32mLast Commit:\033[0m 2025.03.05")
    print(f"\033[1;32mGit Hash:\033[0m 7bdaa90da15a5bfca6a831e739ebdd67fca22299")
    print(f"\033[1;32mContact:\033[0m support@pwmat.com")
    print(f"\033[1;32mCitation:\033[0m https://github.com/LonxunQuantum/PWMLFF")
    print(f"\033[1;32mManual online:\033[0m http://doc.lonxun.com/PWMLFF/")
    print("\033[1;36m" + "=" * 50 + "\033[0m")  # 青色分隔线
    print("\n\n")



import time
import random

def make_iter_name(iter_index: int) :
    iter_format = "%04d"
    return "iter." + (iter_format % iter_index)

def get_iter_from_iter_name(iter_name: str):
    return int(iter_name.split(".")[1].strip())

def make_train_name(model_index: int):
    train_name = "%03d"
    return "train."+(train_name % model_index)

def make_md_name(md_index:int):
    md_name = "%03d"
    return "md."+(md_name % md_index)

def make_md_sys_name(md_index:int, sys_index:int, len_char:int = None):
    md_name = "%03d"
    sys_name = "%03d" if len_char is None else "%0{}d".format(len_char)
    return "md."+(md_name % md_index)+".sys."+(sys_name % sys_index)

def make_temp_press_name(md_index:int, sys_index:int, temp_index:int, press_index:int, len_char:int = None):
    md_name = "%03d"
    sys_name = "%03d" if len_char is None else "%0{}d".format(len_char)
    p_name = "%03d"
    t_name = "%03d"
    return "md."+(md_name % md_index)+".sys."+(sys_name % sys_index)+\
        ".t."+ (sys_name % temp_index) + ".p."+ (p_name % press_index)

def make_temp_name(md_index:int, sys_index:int, temp_index:int, len_char:int = None):
    md_name = "%03d"
    sys_name = "%03d" if len_char is None else "%0{}d".format(len_char)
    t_name = "%03d"
    return "md."+(md_name % md_index)+".sys."+(sys_name % sys_index)+\
        ".t."+ (sys_name % temp_index)

def make_lmps_name(md_index:int, sys_index:int, lmps_index:int, len_char:int = None):
    md_name = "%03d"
    sys_name = "%03d" if len_char is None else "%0{}d".format(len_char)
    return "md."+(md_name % md_index)+".sys."+(sys_name % sys_index)+\
        ".lmps."+ (sys_name % lmps_index)

def get_sub_md_sys_template_name():
    return "md.*.sys.*/md.*.sys.*"

def get_md_sys_template_name():
    return "md.*.sys.*"

def get_traj_file_name(traj_index:int):
    return "{}_traj.lammps".format(traj_index)

def make_scf_name(scf_index:int):
    scf_name = "%04d"
    return "scf."+(scf_name % scf_index)

def get_seed_by_time():
    # seed = int(time.time())
    # random.seed(seed)
    return random.randint(1,10000)
    


import os
import glob
import subprocess
from pwact.utils.file_operation import check_model_type, search_file_by_shell
from pwact.utils.json_operation import get_parameter, get_required_parameter
from pwact.utils.constant import MODEL_CMD, FORCEFILED, UNCERTAINTY, PWDATA
from pwact.active_learning.user_input.train_param.train_param import InputParam as TrainParam
from pwact.active_learning.user_input.scf_param import SCFParam
class InputParam(object):
    # _instance = None
    def __init__(self, json_dict: dict) -> None:
        self.root_dir = get_parameter("work_dir", json_dict, "./")
        if not os.path.isabs(self.root_dir):
            self.root_dir = os.path.realpath(self.root_dir)
        self.record_file = "al.record" #get_parameter("record_file", json_dict, "al.record")
        if "record_file" not in json_dict.keys():
            print("Warning! record_file not provided, automatically set to {}! ".format(self.record_file))
        
        self.reserve_work = get_parameter("reserve_work", json_dict, False)
        # self.reserve_feature = get_parameter("reserve_feature", json_dict, False)
        self.reserve_md_traj = get_parameter("reserve_md_traj", json_dict, False)   #
        self.reserve_scf_files = get_parameter("reserve_scf_files", json_dict, False) # not used

        self.data_format = get_parameter("data_format", json_dict, "extxyz")
        init_data = get_parameter("init_data", json_dict, [])
        self.init_data = self.get_init_data(init_data)
        init_valid_data= get_parameter("valid_data", json_dict, [])
        self.valid_data = self.get_init_data(init_valid_data)
        # the init data for pretraining
        # self.init_data_only_pretrain = get_parameter("init_data_only_pretrain", json_dict, False)
        self.train = TrainParam(json_input=json_dict["train"], cmd=MODEL_CMD.train)
        self.use_pre_model = get_parameter("use_pre_model", json_dict, True)
        self.strategy = StrategyParam(json_dict["strategy"])
        #check_model_type: check type and nums
        self.init_model_list = get_parameter("init_model_list", json_dict, [])
        if len(self.init_model_list) > 0:
            if len(self.init_model_list) != self.strategy.model_num:
                raise Exception("Error! The number of input models needs to be consistent with model_num {} in 'strategy'".format(self.strategy.model_num))
            for _model_file in self.init_model_list:
                if not os.path.exists(_model_file):
                    raise Exception("Error! The model in init_model_list {} does not exist".format(_model_file)) 
                _model_type = check_model_type(_model_file)
                if _model_type != self.train.model_type:
                    raise Exception("Error! The model type in init_model_list is {}, should be consistent with model_type {} in 'train'".format(_model_type, self.train.model_type)) 
            self.init_model_list = [os.path.abspath(_) for _ in self.init_model_list]

        if self.strategy.uncertainty == UNCERTAINTY.kpu and \
            self.train.optimizer_param.opt_name.upper() != "LKF":
            raise Exception("Error! The uncertainty kpu only support the optimizer LKF, please set the 'optimizer/optimizer' in train dict to 'LKF' ")

        self.explore = ExploreParam(json_dict["explore"], self.strategy.max_select)
        self.dft_style = get_required_parameter("dft_style", json_dict["dft"])
        self.scf = SCFParam(json_dict=json_dict["dft"], dft_style=self.dft_style, is_scf=True, root_dir = self.root_dir)

    def to_dict(self):
        res = {}
        res["work_dir"] = self.root_dir
        res["record_file"] = self.record_file
        
        res["reserve_work"] = self.reserve_work
        res["reserve_md_traj"] = self.reserve_md_traj
        res["reserve_scf_files"] = self.reserve_scf_files
        
        res["train"] = self.train.to_dict()
        res["strategy"] = self.strategy.to_dict()

        return res

    def get_init_data(self, init_data:list[str]):
        init_data_path = []
        for _data in init_data:
            _data_path = _data if os.path.isabs(_data) else os.path.join(self.root_dir, _data)
            if not os.path.exists(_data_path):
                print("Warning! the init data {} does not exist!".format(_data_path))
                continue
            init_data_path.append(_data_path)
        return init_data_path
        
class StrategyParam(object):
    def __init__(self, json_dict) -> None:
        self.md_type = get_parameter("md_type", json_dict, FORCEFILED.libtorch_lmps)
        self.lmps_tolerance = get_parameter("lmps_tolerance", json_dict, True)
        self.max_select = get_parameter("max_select", json_dict, None)
        self.uncertainty = get_parameter("uncertainty", json_dict, UNCERTAINTY.committee).upper()
        if self.uncertainty.upper() == UNCERTAINTY.kpu:
            self.model_num = 1
            self.kpu_upper = get_parameter("kpu_upper", json_dict, 1.5)
            self.kpu_lower = get_parameter("kpu_lower", json_dict, 0.5)
        elif self.uncertainty.upper() == UNCERTAINTY.committee:
            self.model_num = get_parameter("model_num", json_dict, 4)
            if self.model_num < 3:
                raise Exception("ERROR! for {}, make sure model_num >= 3".format(UNCERTAINTY.committee))

            self.lower_model_devi_f = get_parameter("lower_model_deiv_f", json_dict, None)#  lower_model_deiv_f 早期版本手误错写为deiv_f, 这里做兼容性处理
            self.upper_model_devi_f = get_parameter("upper_model_deiv_f", json_dict, None)
            if self.lower_model_devi_f is None:
                self.lower_model_devi_f = get_required_parameter("lower_model_devi_f", json_dict) 
            if self.upper_model_devi_f is None:
                self.upper_model_devi_f = get_required_parameter("upper_model_devi_f", json_dict)

        else:
            raise Exception("ERROR! uncertainty strategy {} not support! Please check!".format(self.uncertainty))

        self.compress = get_parameter("compress", json_dict, False)
        self.compress_order = get_parameter("compress_order", json_dict, 3)
        self.compress_dx = get_parameter("compress_dx", json_dict, 0.01)
        
        if self.compress:
            if self.uncertainty == UNCERTAINTY.committee and self.md_type == FORCEFILED.fortran_lmps:
                raise Exception("Error! The compress model does not fortran lammps! Please set the 'md_type' to 2!")
        if self.uncertainty == UNCERTAINTY.kpu:
            if self.compress:
                error_log = "Error! the kpu uncertainty does not support compress, please set the 'compress' in strategy dict to be false!"
                raise Exception(error_log)
        
        self.direct = get_parameter("direct", json_dict, False)
        if self.direct:
            self.direct_script = get_parameter("direct_script", json_dict, None)
            if self.direct_script is not None:
                self.direct_script = os.path.abspath(self.direct_script)
                if not os.path.exists(self.direct_script):
                    raise Exception("ERROR! The direct script {} does not exist!".format(self.direct_script))
            else:
                raise Exception("ERROR! The direct script does not exist!")
        else:
            self.direct_script = None

    def to_dict(self):
        res = {}
        res["md_type"] = self.md_type
        res["max_select"] = self.max_select
        res["uncertainty"] = self.uncertainty
        res["model_num"] = self.model_num

class SysConfig(object):
    '''
    description: 
        for wildcard such as "scale-1.000/00000*/POSCAR"
    param {*} self
    param {str} sys_config
    param {str} format
    return {*}
    author: wuxingxing
    '''    
    def __init__(self, sys_config:str, format:str) -> None:
        # self.sys_config = sys_config
        self.format = format
        sys_config_list = glob.glob(sys_config)
        self.sys_config = sorted(sys_config_list)

class ExploreParam(object):
    def __init__(self, json_dict, max_select:int=None) -> None:
        # sys_configs
        sys_config_prefix = get_parameter("sys_config_prefix", json_dict, None)
        sys_configs = get_required_parameter("sys_configs", json_dict)
        if isinstance(sys_configs, str) or isinstance(sys_configs, dict):
            sys_configs = [sys_configs]
        self.sys_configs:list[SysConfig]=[]
        # self.sys_configs[0] = config files with * and format
        for sys_config in sys_configs:
            if isinstance(sys_config, str):
                config = os.path.join(sys_config_prefix, sys_config) if sys_config_prefix is not None else sys_config
                config_format = PWDATA.pwmat_config
            elif isinstance(sys_config, dict):
                config = os.path.join(sys_config_prefix, sys_config["config"]) if sys_config_prefix is not None else sys_config["config"]
                config_format = get_parameter("format", sys_config, PWDATA.pwmat_config)
            if len(glob.glob(config)) < 1:
                raise Exception("ERROR! The sys_config {} file does not exist!".format(config))
            self.sys_configs.append(SysConfig(config, config_format))

        # lammps.in files
        lmps_prefix = get_parameter("lmps_prefix", json_dict, None)
        lmps_in = get_parameter("lmps_in", json_dict, [])
        if isinstance(lmps_in, str) or isinstance(lmps_in, dict):
            lmps_in = [lmps_in]
        self.lmps_in:list[str]=[]
        for lmp_in_file in lmps_in:
            lmp_file = os.path.join(lmps_prefix, lmp_in_file) if lmps_prefix is not None else lmp_in_file
            if not os.path.exists(lmp_file):
                raise Exception("ERROR! The lammps.in file {} does not exist!".format(lmp_file))
            self.lmps_in.append(lmp_file)
        
        # set md deatils
        self.md_job_list = self.set_md_details(json_dict["md_jobs"], max_select)
        self.md_job_num = len(self.md_job_list)

    def set_md_details(self, md_list_dict:list[dict], max_select):
        iter_md:list[list[MdDetail]] = []
        for iter_index, md_dict in enumerate(md_list_dict): # for each iter
            iter_exp_md:list[MdDetail] = []
            if not isinstance(md_dict, list):
                md_dict = [md_dict]
            for md_exp_id, md_exp in enumerate(md_dict):
                iter_exp_md.append(MdDetail(md_exp_id, md_exp, max_select, self.sys_configs, self.lmps_in))
            iter_md.append(iter_exp_md)
        return iter_md
    
    def to_dict(self):
        res = {}
        return res

class MdDetail(object):
    def __init__(self, md_index: int, 
                        json_dict:dict, 
                        max_select:int=None, 
                        sys_configs:list[SysConfig]=None,
                        lmps_in:list[str]=None) -> None:
        self.md_index = md_index
        self.trj_freq = get_parameter("trj_freq", json_dict, 10)

        self.nsteps = get_parameter("nsteps", json_dict, None)
        self.md_dt = get_parameter("md_dt", json_dict, 0.001) #fs
        
        self.ensemble = get_parameter("ensemble", json_dict, "nve")
        
        self.press_list = get_parameter("press", json_dict, [])
        self.taup = get_parameter("taup", json_dict, 0.5)
        if not isinstance(self.press_list, list):
            self.press_list = [self.press_list]
        self.temp_list = get_parameter("temps", json_dict, [])
        self.taut = get_parameter("taut", json_dict, 0.1)
        
        if not isinstance(self.temp_list, list):
            self.temp_list = [self.temp_list]

        #sys_idx
        sys_idx = get_required_parameter("sys_idx", json_dict)

        if not isinstance(sys_idx, list):
            self.sys_idx = [sys_idx]
        _select_sys = get_parameter("select_sys", json_dict, None)
        if _select_sys is None:
            if max_select is None:
                _select_sys = 100 # if the max_select and select_sys are all None, set the select_sys to 100 as default
            else:
                _select_sys = max_select
        if not isinstance(_select_sys, list):
            _select_sys = [_select_sys]
        
        #select_sys
        select_sys = []
        if len(_select_sys) > 0:
            if len(_select_sys) == 1:
                for i in range(0, len(sys_idx)):
                    select_sys.append(_select_sys[0])
            elif len(_select_sys) == len(sys_idx):
                select_sys = _select_sys
            else:
                raise Exception("The length of the 'select_sys' array needs to be consistent with'sys_idx'" )
        
        # from lammps.in
        _lmps_in_idx = get_parameter("lmps_in_idx", json_dict, [])
        lmps_in_idx = []
        if not isinstance(_lmps_in_idx, list):
            _lmps_in_idx = [_lmps_in_idx]
        # check lammps.in file
        if len(_lmps_in_idx) > 0:
            if len(_lmps_in_idx) == 1:
                for i in range(0, len(sys_idx)):
                    lmps_in_idx.append(_lmps_in_idx[0])
            elif len(_lmps_in_idx) == len(sys_idx):
                lmps_in_idx = _lmps_in_idx
            else:
                raise Exception("The length of the 'lmps_in_idx' array needs to be consistent with'sys_idx'" )
            self.use_lmps_in = True
        else:
            self.use_lmps_in = False

        # reset select_sys and sys_idx by sys_configs
        self.sys_idx = []
        self.select_sys = []
        self.lmp_in_idx = []
        self.config_file_list = []
        self.lmp_in_file_list = []
        self.config_file_format = []
        file_id = 0
        for index, sys_id in enumerate(sys_idx):
            systems = sys_configs[sys_id].sys_config
            if self.use_lmps_in:
                lmp_in_file = lmps_in[lmps_in_idx[index]]
            system_format = sys_configs[sys_id].format
            for system in systems:
                self.config_file_list.append(system)
                self.config_file_format.append(system_format)
                self.sys_idx.append(file_id)
                self.select_sys.append(select_sys[index])
                if self.use_lmps_in:
                    self.lmp_in_idx.append(lmps_in_idx[index])
                    self.lmp_in_file_list.append(lmp_in_file)
                file_id += 1

        self.kspacing = get_parameter("temps", json_dict, None)
        self.neigh_modify = get_parameter("neigh_modify", json_dict, 10)
        self.mass = get_parameter("mass",json_dict, None)
        self.merge_traj = get_parameter("merge_traj", json_dict, False)
        self.boundary = get_parameter("boundary", json_dict, True)
    '''
    description: 
    maybe not use
    param {*} self
    return {*}
    author: wuxingxing
    '''    
    def get_md_combination(self):
        md_list = []
        for t, temp in enumerate(self.temp_list):
            for p, press in enumerate(self.press_list):
               md_list.append(t, p) 
        



from ase.io import read
from ase.units import fs
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution
from ase.md.nvtberendsen import NVTBerendsen
from ase.md.npt import NPT
from ase.optimize import LBFGS
import numpy as np
from sevenn.sevennet_calculator import SevenNetCalculator

traj_name = "tmp.traj"
xyz_name = "traj.xyz"

calc = SevenNetCalculator()
fmax = 0.1
run_step = 10000

T = 300
P = 1.01325
P_in_ev_per_ang3 = P / 1602176.6208
atoms = read("POSCAR")
if not NPT._isuppertriangular(atoms.get_cell()):
    a, b, c, alpha, beta, gamma = atoms.cell.cellpar()
    angles = np.radians((alpha, beta, gamma))
    sin_a, sin_b, _sin_g = np.sin(angles)
    cos_a, cos_b, cos_g = np.cos(angles)
    cos_p = (cos_g - cos_a * cos_b) / (sin_a * sin_b)
    cos_p = np.clip(cos_p, -1, 1)
    sin_p = (1 - cos_p**2) ** 0.5
    new_basis = [
        (a * sin_b * sin_p, a * sin_b * cos_p, a * cos_b),
        (0, b * sin_a, b * cos_a),
        (0, 0, c),
    ]
    atoms.set_cell(new_basis, scale_atoms=True)
    
atoms.calc = calc
opt = LBFGS(atoms)

def atoms2xyzstr(atoms):
    num_atom = atoms.get_global_number_of_atoms()
    vol = atoms.get_volume()
    pos = atoms.positions
    forces = atoms.get_forces()
    energy = atoms.get_potential_energy()
    cell = atoms.cell
    virial = -atoms.get_stress(voigt=False) * vol
    xyzstr = "%d\n" % num_atom
    xyz_head = 'Lattice="%.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f" Properties=species:S:1:pos:R:3:forces:R:3 energy=%.8f' 
    xyz_format = (cell[0,0],cell[0,1],cell[0,2],cell[1,0],cell[1,1],cell[1,2],cell[2,0],cell[2,1],cell[2,2],energy)
    if virial is not None:
        xyz_head += ' virial="%.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f"'
        xyz_format += (
            virial[0,0], virial[0,1], virial[0,2],
            virial[1,0], virial[1,1], virial[1,2],
            virial[2,0], virial[2,1], virial[2,2]
            )
    xyz_head += '\n'
    xyzstr += xyz_head % xyz_format
    for i in range(num_atom):
        xyzstr += "%2s %14.8f %14.8f %14.8f %14.8f %14.8f %14.8f\n" %\
        (atoms[i].symbol,pos[i,0],pos[i,1],pos[i,2],forces[i,0],forces[i,1],forces[i,2])
    return xyzstr

def domd():
    opt.run(fmax=fmax)
    MaxwellBoltzmannDistribution(atoms, temperature_K=T)

    '''
    print("Beginning Nose-Hoover NVT equilibration")
    dyn_nvt = NPT(
    atoms=atoms,
    timestep=.5*fs,
    temperature_K=600,
    externalstress=P_in_ev_per_ang3,
    ttime=25*fs,
    pfactor=None,
    logfile="nvt.log",
    loginterval=200
    )
    dyn_nvt.run(200000)
    '''

    print("Beginning Nose-Hoover NPT equilibration")
    dyn_npt = NPT(
    atoms=atoms,
    timestep=.5*fs,
    temperature_K=T,
    externalstress=P_in_ev_per_ang3,
    mask=np.eye(3),
    ttime=25*fs,
    pfactor=50*fs, #None nvt
    trajectory=traj_name,
    logfile="npt.log",
    loginterval=100
    )
    dyn_npt.run(run_step)

def dolabel():
    traj = read(traj_name, index=":")
    output_file = xyz_name
    f = open(output_file, "w")
    for i in range(len(traj)):
        atoms = traj[i]
        atoms.calc = calc
        f.write(atoms2xyzstr(atoms))
    f.close()

if __name__=="__main__":
    domd()
    dolabel()


#!/usr/bin/env python3
import re
import random
import os
import numpy as np

from pwact.utils.constant import FORCEFILED
from pwdata.calculators.const import ELEMENTMASSTABLE
def _sample_sphere():
    while True:
        vv = np.array([np.random.normal(), np.random.normal(), np.random.normal()])
        vn = np.linalg.norm(vv)
        if vn < 0.2:
            continue
        return vv / vn

def make_pair_style(md_type, forcefiled, atom_type:list[int], dump_info:str, matpl): # matpl/dp matpl/nep or matpl/nep/kk
    pair_style = ""
    
    if md_type == FORCEFILED.fortran_lmps:
        raise Exception("the fortran in.lammps not relized!")
        pass
    
    elif md_type == FORCEFILED.libtorch_lmps:
        pair_names = ""
        for fi in forcefiled:
            pair_names += "{} ".format(os.path.basename(fi))
        pair_style = "pair_style   {}  {} {}\n".format(matpl, pair_names, dump_info)
    atom_names = " ".join(map(str, atom_type))
    pair_style += "pair_coeff       * * {}\n".format(atom_names)
    return pair_style

# def make_mass(mass):
#     if mass is None:
#         return ""
    
#     mass_str = "\n"
#     if isinstance(mass, str):
#         mass_str += "mass       {}\n".format(mass)
#     else:
#         for m in mass:
#             mass_str += "mass       {}\n".format(m)
#     return mass_str

def make_mass(atom_type:list):
    if isinstance(atom_type, str):
        atom_type = [atom_type]
    if atom_type is None:
        return ""
    mass_str = "\n"
    
    for i, atom in enumerate(atom_type):
        mass_str += "mass   {}    {}\n".format(i+1, ELEMENTMASSTABLE[atom])
    return mass_str

def set_ensemble(ensemble):
    pass

def make_lammps_input(
    md_file,
    md_type,
    forcefiled,
    atom_type,
    ensemble,
    nsteps,
    dt,
    neigh_modify,
    trj_freq,
    mass,
    temp,
    tau_t,
    press,
    tau_p,    
    boundary, #true is 'p p p', false is 'f f f', default is true
    merge_traj,
    max_seed=100000,
    restart=0,
    model_deviation_file = "model_deviation.out",
    kokkos_head=False,
    matpl="matpl/nep"
):
    if kokkos_head:
        md_script = "package kokkos neigh half comm device\n"
        md_script += "newton on\n"
    else:
        md_script = ""
    md_script += "variable        NSTEPS          equal %d\n" % nsteps
    md_script += "variable        THERMO_FREQ     equal %d\n" % trj_freq
    md_script += "variable        DUMP_FREQ       equal %d\n" % trj_freq
    md_script += "variable        restart         equal %d\n" % restart

    md_script += "variable        TEMP            equal %f\n" % temp
    # if ele_temp_f is not None:
    #     md_script += "variable    ELE_TEMP        equal %f\n" % ele_temp_f
    # if ele_temp_a is not None:
    #     md_script += "variable    ELE_TEMP        equal %f\n" % ele_temp_a
    if press is not None:
        md_script += "variable        PRESS           equal %f\n" % press
    md_script += "variable        TAU_T           equal %f\n" % tau_t
    if tau_p is not None:
        md_script += "variable        TAU_P           equal %f\n" % tau_p
    md_script += "\n"
    
    md_script += "units           metal\n"

    if boundary:
        md_script += "boundary        p p p\n"
    else:
        md_script += "boundary        f f f\n"
    
    md_script += "atom_style      atomic\n"
    md_script += "\n"
    md_script += "neighbor        1.0 bin\n"
    if neigh_modify is not None:
        md_script += "neigh_modify    delay %d\n" % neigh_modify
    md_script += "\n"
    md_script += "box              tilt large\n"
    
    md_script += (
        'if "${restart} > 0" then "read_restart lmps.restart.*" else "read_data %s"\n'
        % md_file
    )
    md_script += "change_box       all triclinic\n"
    
    md_script += make_mass(atom_type)
    dump_info = "out_freq ${{DUMP_FREQ}} out_file {} ".format(model_deviation_file)
    md_script += make_pair_style(md_type, forcefiled, atom_type, dump_info, matpl)
    #put_freq ${freq} out_file error
    
    md_script += "\n"
    md_script += "thermo_style    custom step temp pe ke etotal press vol lx ly lz xy xz yz\n"
    md_script += "thermo          ${THERMO_FREQ}\n"
    
    if merge_traj is True:
        md_script += "dump            1 all custom ${DUMP_FREQ} all.lammpstrj id type x y z fx fy fz\n"
        md_script += 'if "${restart} > 0" then "dump_modify     1 append yes"\n'
    else:
        md_script += "dump            1 all custom ${DUMP_FREQ} traj/*.lammpstrj id type x y z fx fy fz\n"
    
    md_script += "restart         10000 lmps.restart\n"
    md_script += "\n"
    
    md_script += 'if "${restart} == 0" then "velocity        all create ${TEMP} %d"' % (
            random.randrange(max_seed - 1) + 1
        )
        
    md_script += "\n"
    
    if ensemble.split("-")[0] == "npt":
        assert press is not None
        if not boundary:
            raise RuntimeError("ensemble %s is conflicting with boundary" % ensemble)
    if ensemble == "npt" or ensemble == "npt-i" or ensemble == "npt-iso":
        md_script += "fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRESS} ${PRESS} ${TAU_P}\n"
    elif ensemble == "npt-a" or ensemble == "npt-aniso":
        md_script += "fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} aniso ${PRESS} ${PRESS} ${TAU_P}\n"
    elif ensemble == "npt-t" or ensemble == "npt-tri":
        md_script += "fix             1 all npt temp ${TEMP} ${TEMP} ${TAU_T} tri ${PRESS} ${PRESS} ${TAU_P}\n"
    elif ensemble == "nvt":
        md_script += "fix             1 all nvt temp ${TEMP} ${TEMP} ${TAU_T}\n"
    elif ensemble == "nve":
        md_script += "fix             1 all nve\n"
    else:
        raise RuntimeError("unknown emsemble " + ensemble)
    
    if not boundary:
        md_script += "velocity        all zero linear\n"
        md_script += "fix             fm all momentum 1 linear 1 1 1\n"
    md_script += "\n"
    md_script += "timestep        %f\n" % dt
    md_script += "run             ${NSTEPS} upto\n"
    
    return md_script

def get_dumped_forces(file_name):
    with open(file_name) as fp:
        lines = fp.read().split("\n")
    natoms = None
    for idx, ii in enumerate(lines):
        if "ITEM: NUMBER OF ATOMS" in ii:
            natoms = int(lines[idx + 1])
            break
    if natoms is None:
        raise RuntimeError(
            "wrong dump file format, cannot find number of atoms", file_name
        )
    idfx = None
    for idx, ii in enumerate(lines):
        if "ITEM: ATOMS" in ii:
            keys = ii
            keys = keys.replace("ITEM: ATOMS", "")
            keys = keys.split()
            idfx = keys.index("fx")
            idfy = keys.index("fy")
            idfz = keys.index("fz")
            break
    if idfx is None:
        raise RuntimeError("wrong dump file format, cannot find dump keys", file_name)
    ret = []
    for ii in range(idx + 1, idx + natoms + 1):
        words = lines[ii].split()
        ret.append([float(words[ii]) for ii in [idfx, idfy, idfz]])
    ret = np.array(ret)
    return ret


def get_all_dumped_forces(file_name):
    with open(file_name) as fp:
        lines = fp.read().split("\n")

    ret = []
    exist_natoms = False
    exist_atoms = False

    for idx, ii in enumerate(lines):

        if "ITEM: NUMBER OF ATOMS" in ii:
            natoms = int(lines[idx + 1])
            exist_natoms = True

        if "ITEM: ATOMS" in ii:
            keys = ii
            keys = keys.replace("ITEM: ATOMS", "")
            keys = keys.split()
            idfx = keys.index("fx")
            idfy = keys.index("fy")
            idfz = keys.index("fz")
            exist_atoms = True

            single_traj = []
            for jj in range(idx + 1, idx + natoms + 1):
                words = lines[jj].split()
                single_traj.append([float(words[jj]) for jj in [idfx, idfy, idfz]])
            single_traj = np.array(single_traj)
            ret.append(single_traj)

    if exist_natoms is False:
        raise RuntimeError(
            "wrong dump file format, cannot find number of atoms", file_name
        )
    if exist_atoms is False:
        raise RuntimeError("wrong dump file format, cannot find dump keys", file_name)
    return ret


def make_lammps_input_from_lmp_in_file(
    md_file,
    md_type,
    forcefiled,
    lmp_in_file,
    atom_type,
    trj_freq,
    boundary, #true is 'p p p', false is 'f f f', default is true
    merge_traj,
    max_seed=100000,
    restart=0,
    model_deviation_file = "model_deviation.out",
    matpl="matpl/nep"
):
    with open(lmp_in_file, 'r') as rf:
        lmp_content = rf.readlines()
    # 如果存在 kokkos 设置在去掉之后追加到文件头
    lmp_content, has_kokos = remove_package_and_newton_lines(lmp_content)

    runline_idx = find_first_run_cmd_line(lmp_content)
    if runline_idx is None:
        raise Exception("Error! The input lmp.in file: {} is missing the 'RUN' command, please modify it!".format(lmp_in_file))
    # remove the units boundary atom_stype lines
    # units           metal
    # boundary        p p p
    # atom_style      atomic
    # remove mass pair_style pair_coeff dump
    lmp_content, insert_info = remove_lmps_lines(lmp_content)
    lmp_content.insert(0, "variable        DUMP_FREQ       equal %d\n" % trj_freq)
    lmp_content.insert(1, "variable        restart         equal %d\n" % restart)
    
    md_script = ""
    md_script += "units           metal\n"
    if boundary:
        md_script += "boundary        p p p\n"
    else:
        md_script += "boundary        f f f\n"
    md_script += "atom_style      atomic\n"
    md_script += "\n"
    lmp_content.insert(2, md_script)
    
    md_script = (
        'if "${restart} > 0" then "read_restart lmps.restart.*" else "read_data %s"\n'
        % md_file
    )
    lmp_content.insert(3, md_script)
    
    md_script = make_mass(atom_type)
    dump_info = "out_freq ${{DUMP_FREQ}} out_file {} ".format(model_deviation_file)
    md_script += make_pair_style(md_type, forcefiled, atom_type, dump_info, matpl)
    #put_freq ${freq} out_file error
    md_script += "\n"
    lmp_content.insert(4, md_script)
    # md_script += "thermo_style    custom step temp pe ke etotal press vol lx ly lz xy xz yz\n"
    # md_script += "thermo          ${THERMO_FREQ}\n"
    
    if merge_traj is True:
        dump_line = "dump            1 all custom ${DUMP_FREQ} all.lammpstrj id type x y z fx fy fz\n"
        if "dump_modify" in insert_info.keys():
            dump_line += 'if "${restart} > 0" then' + 'dump_modify 1 {} append yes\n'.format(" ".join(insert_info['dump_modify'].split()[2:]))
        else:
            dump_line += 'if "${restart} > 0" then "dump_modify     1 append yes"\n'
    else:
        dump_line = "dump            1 all custom ${DUMP_FREQ} traj/*.lammpstrj id type x y z fx fy fz\n"
        if "dump_modify" in insert_info.keys():
            dump_line += 'dump_modify 1 {}\n'.format(" ".join(insert_info['dump_modify'].split()[2:]))

    dump_line += "restart         10000 lmps.restart\n"
    dump_line += "\n"
    
    dump_line += 'if "${restart} == 0" then "velocity        all create ${TEMP} %d"' % (
            random.randrange(max_seed - 1) + 1
        )
        
    dump_line += "\n"
    lmp_content.insert(runline_idx, dump_line)

    if has_kokos:
        lmp_content.insert(0, "package kokkos neigh half comm device\n") 
        lmp_content.insert(1, "newton on\n")
    return "".join(lmp_content)

def remove_lmps_lines(lmps_lines):
    removes = ["dump_freq", "restart","read_data"]
    remove_head = ["mass", "units", "boundary", "atom_style", "pair_style", "pair_coeff", "dump"]

    insert_head = ["dump_modify"]
    new_line = []
    insert_info = {}
    for idx, line in enumerate(lmps_lines):
        if line.lstrip() == '' or line.lstrip().startswith('#'):
            new_line.append(line)
            continue
        if any (keyword.lower() == line.lstrip().split()[0].lower() for keyword in insert_head):
            insert_info[line.lstrip().split()[0].lower()] = line
            continue
        if any(
            keyword.lower() in line.lower() for keyword in removes
        ):
            continue
        if any (keyword.lower() == line.lstrip().split()[0].lower() for keyword in remove_head):
            continue
        new_line.append(line)
    # new_lines = [
    #     line for line in lmps_lines 
    #     if line.lstrip().startswith('#')
    #     or (not any(
    #         keyword.lower() in line.lower() 
    #         for keyword in removes
    #     ) and not (
    #         keyword.lower() in line.lstrip().split()[0].lower() 
    #         for keyword in remove_head
    #     ))
    #     ]
    return new_line, insert_info


def remove_package_and_newton_lines(lines):
    package_pattern = re.compile(r'package\s+kokkos\s+neigh', re.IGNORECASE)
    newton_pattern = re.compile(r'newton\s+on', re.IGNORECASE)

    has_package = any(package_pattern.search(line) for line in lines)
    if not (has_package):
        return lines, False
    
    new_lines = []
    for line in lines:
        if package_pattern.search(line) or newton_pattern.search(line):
            continue
        new_lines.append(line)
    return new_lines, True


def find_first_run_cmd_line(lmps_lines):
    for i, line in enumerate(lmps_lines):
        if 'run' in line.lower():
            # 计算倒数行数(从1开始计数)
            return -(len(lmps_lines) - i)
    return None

if __name__=="__main__":
    lmp_in_file = "/data/home/wuxingxing/datas/debugs/czy/lammps.in"
    with open(lmp_in_file, 'r') as rf:
        lmp_content = rf.readlines()
    runline_idx = find_first_run_cmd_line(lmp_content)
    if runline_idx is None:
        raise Exception("Error! The input lmp.in file: {} is missing the 'RUN' command, please modify it!".format(lmp_in_file))
    # remove the units boundary atom_stype lines
    # units           metal
    # boundary        p p p
    # atom_style      atomic
    # remove mass pair_style pair_coeff dump
    lmp_content,insert_info = remove_lmps_lines(lmp_content)
    merge_traj = False
    if merge_traj is True:
        dump_line = "dump            1 all custom ${DUMP_FREQ} all.lammpstrj id type x y z fx fy fz\n"
        if "dump_modify" in insert_info.keys():
            dump_line += 'if "${restart} > 0" then' + 'dump_modify 1 {} append yes\n'.format(" ".join(insert_info['dump_modify'].split()[2:]))
        else:
            dump_line += 'if "${restart} > 0" then "dump_modify     1 append yes"\n'
    else:
        dump_line = "dump            1 all custom ${DUMP_FREQ} traj/*.lammpstrj id type x y z fx fy fz\n"
        if "dump_modify" in insert_info.keys():
            dump_line += 'dump_modify 1 {}\n'.format(" ".join(insert_info['dump_modify'].split()[2:]))


from pwact.utils.json_operation import get_parameter, get_required_parameter
from pwact.active_learning.user_input.train_param.nep_param import NepParam
class OptimizerParam(object):
    def __init__(self) -> None:
        pass

    def set_optimizer(self, json_source:dict, nep_param:NepParam=None):
        optimizer_dict = get_parameter("optimizer", json_source, {})
        self.opt_name = get_parameter("optimizer", optimizer_dict, "ADAM")
        self.batch_size = get_parameter("batch_size", optimizer_dict, 1)
        self.epochs = get_parameter("epochs", optimizer_dict, 30)
        self.print_freq = get_parameter("print_freq", optimizer_dict, 10)
        # the start epoch could be reset at the resume model code block
        self.reset_epoch = get_parameter("reset_epoch", optimizer_dict, False)
        self.start_epoch = get_parameter("start_epoch", optimizer_dict, 1)

        self.lambda_1 = None
        if "lambda_1" in optimizer_dict:
            raise Warning("The lambda_1 is not realized now!")
            # self.lambda_1 = get_parameter("lambda_1", optimizer_dict, None) # weight of regularization term
            # if self.lambda_1 is not None and self.lambda_1 < 0:
            #     raise Exception("ERROR! the lambda_1 should >= 0 !")

        self.lambda_2 = get_parameter("lambda_2", optimizer_dict, None) # weight of norm regularization term
        if self.lambda_2 is not None and self.lambda_2 < 0:
            raise Exception("ERROR! the lambda_2 should >= 0 !")

        if "KF" in self.opt_name.upper():  #set Kalman Filter Optimizer params
            self.kalman_lambda = get_parameter("kalman_lambda", optimizer_dict, 0.98)
            self.kalman_nue = get_parameter("kalman_nue", optimizer_dict, 0.9987)
            self.block_size = get_parameter("block_size", optimizer_dict, 5120)
            self.nselect = get_parameter("nselect", optimizer_dict, 24)
            self.groupsize = get_parameter("groupsize", optimizer_dict, 6)
            self.p0_weight = get_parameter("p0_weight", optimizer_dict, None)

        elif "ADAM" in self.opt_name.upper():   # set ADAM Optimizer params
            self.learning_rate = get_parameter("learning_rate", optimizer_dict, 0.001)
            self.weight_decay = get_parameter("weight_decay", optimizer_dict, 1e-4)
            self.momentum = get_parameter("momentum", optimizer_dict, 0.9)
            self.gamma = get_parameter("gamma", optimizer_dict, 0.99) # used in nn optimizer
            self.step = get_parameter("step", optimizer_dict, 100) # used in nn optimizer
            self.scheduler = get_parameter("scheduler", optimizer_dict, None) # used in nn optimizer

            self.stop_step = get_parameter("stop_step", optimizer_dict, 1000000)
            self.decay_step = get_parameter("decay_step", optimizer_dict, 5000)
            self.stop_lr = get_parameter("stop_lr",optimizer_dict, 3.51e-8)
            # self.set_adam_sgd_params(learning_rate, weight_decay, momentum,\
            #                                 gamma, step, scheduler, stop_step, decay_step, stop_lr)
        else:
            pass

        self.train_energy = get_parameter("train_energy", optimizer_dict, True) 
        self.train_force = get_parameter("train_force", optimizer_dict, True) 
        self.train_ei = get_parameter("train_ei", optimizer_dict, False) 
        self.train_virial = get_parameter("train_virial", optimizer_dict, False) 
        self.train_egroup = get_parameter("train_egroup", optimizer_dict, False) 

        self.force_delta = None
        self.population = None
        self.generation = None
        self.pre_fac_force = 2.0
        self.pre_fac_etot = 1.0
        self.pre_fac_ei = 1.0
        self.pre_fac_virial = 1.0
        self.pre_fac_egroup = 0.1

        if "KF" in self.opt_name.upper():
            self.pre_fac_force = get_parameter("pre_fac_force", optimizer_dict, 2.0) 
            self.pre_fac_etot = get_parameter("pre_fac_etot", optimizer_dict, 1.0) 
            self.pre_fac_ei = get_parameter("pre_fac_ei", optimizer_dict, 1.0) 
            self.pre_fac_virial = get_parameter("pre_fac_virial", optimizer_dict, 1.0) 
            self.pre_fac_egroup = get_parameter("pre_fac_egroup", optimizer_dict, 0.1) 
            
        elif "ADAM" in self.opt_name.upper():
            self.start_pre_fac_force = get_parameter("start_pre_fac_force", optimizer_dict, 1000) 
            self.start_pre_fac_etot = get_parameter("start_pre_fac_etot", optimizer_dict, 0.02) 
            self.start_pre_fac_ei = get_parameter("start_pre_fac_ei", optimizer_dict, 0.1) 
            self.start_pre_fac_virial = get_parameter("start_pre_fac_virial", optimizer_dict, 50.0) 
            self.start_pre_fac_egroup = get_parameter("start_pre_fac_egroup", optimizer_dict, 0.02) 

            self.end_pre_fac_force = get_parameter("end_pre_fac_force", optimizer_dict, 1.0) 
            self.end_pre_fac_etot = get_parameter("end_pre_fac_etot", optimizer_dict, 1.0) 
            self.end_pre_fac_ei = get_parameter("end_pre_fac_ei", optimizer_dict, 2.0) 
            self.end_pre_fac_virial = get_parameter("end_pre_fac_virial", optimizer_dict, 1.0) 
            self.end_pre_fac_egroup = get_parameter("end_pre_fac_egroup", optimizer_dict, 1.0) 

        elif "SNES" in self.opt_name.upper():# natural evolution strategies
            # if get_parameter("nep_in_file", json_source, None) is not None \
            #     or get_parameter("nep_txt_file", json_source, None) is not None:
            try: # read from nep.in file
                self.pre_fac_etot = nep_param.lambda_e
                self.pre_fac_force = nep_param.lambda_f
                self.pre_fac_virial = nep_param.lambda_v
                self.pre_fac_egroup = nep_param.lambda_eg
                self.pre_fac_ei = nep_param.lambda_ei
                self.lambda_1 = nep_param.lambda_1
                self.lambda_2 = nep_param.lambda_2
                self.force_delta = nep_param.force_delta
                self.population = nep_param.population
                self.generation = nep_param.generation
                self.batch_size = nep_param.batch
                self.eta_m = None
                self.eta_s = None
                return

            except Exception:
                print('Read snes optimizer param from json file')
            # from 'optimizer' dict
            self.pre_fac_ei = get_parameter("lambda_ei", optimizer_dict, 1.0) # weight of energy loss term
            self.pre_fac_egroup = get_parameter("lambda_eg", optimizer_dict, 0.1) # weight of energy loss term
            self.pre_fac_etot = get_parameter("lambda_e", optimizer_dict, 1.0) # weight of energy loss term
            self.pre_fac_force = get_parameter("lambda_f", optimizer_dict, 1.0) # weight of force loss term
            self.pre_fac_virial = get_parameter("lambda_v", optimizer_dict, 0.1) # weight of virial loss term
            self.force_delta = get_parameter("force_delta", optimizer_dict, None) # bias term that can be used to make smaller forces more accurate
            self.batch_size = get_parameter("batch_size", optimizer_dict, 1000) # batch size for training
            self.population = get_parameter("population", optimizer_dict, 50) # population size used in the SNES algorithm [Schaul2011]
            self.generation = get_parameter("generation", optimizer_dict, 100000) # number of generations used by the SNES algorithm [Schaul2011]
            self.eta_m = get_parameter("eta_m", optimizer_dict, None) # population size used in the SNES algorithm [Schaul2011]
            self.eta_s = get_parameter("eta_s", optimizer_dict, None) # number of generations used by the SNES algorithm [Schaul2011]
        
    def to_linear_dict(self):
        opt_dict = {}
        opt_dict["train_energy"] = self.train_energy
        opt_dict["train_force"] = self.train_force
        opt_dict["train_ei"] = self.train_ei
        opt_dict["pre_fac_force"] = self.pre_fac_force
        opt_dict["pre_fac_etot"] = self.pre_fac_etot
        opt_dict["pre_fac_ei"] = self.pre_fac_ei
        return opt_dict
    
    def to_dict(self):
        opt_dict = {}
        opt_dict["optimizer"]=self.opt_name
        # opt_dict["start_epoch"] = self.start_epoch
        opt_dict["epochs"] = self.epochs
        opt_dict["batch_size"] = self.batch_size
        opt_dict["print_freq"] = self.print_freq
        if self.lambda_1 is not None:
            opt_dict["lambda_1"] =  self.lambda_1
        if self.lambda_2 is not None:  
            opt_dict["lambda_2"] =  self.lambda_2

        if "KF" in self.opt_name:
            if "LKF" in self.opt_name:
                opt_dict["block_size"] = self.block_size 
            opt_dict["kalman_lambda"] = self.kalman_lambda
            opt_dict["kalman_nue"] = self.kalman_nue

            opt_dict["train_energy"] = self.train_energy
            opt_dict["train_force"] = self.train_force
            opt_dict["train_ei"] = self.train_ei
            opt_dict["train_virial"] = self.train_virial
            opt_dict["train_egroup"] = self.train_egroup
    
            opt_dict["pre_fac_force"] = self.pre_fac_force
            opt_dict["pre_fac_etot"] = self.pre_fac_etot
            opt_dict["pre_fac_ei"] = self.pre_fac_ei
            opt_dict["pre_fac_virial"] = self.pre_fac_virial
            opt_dict["pre_fac_egroup"] = self.pre_fac_egroup

            if self.p0_weight is not None:
                opt_dict["p0_weight"] = self.p0_weight

        elif "SGD" in self.opt_name or "ADAM" in self.opt_name:
            if "SGD" in self.opt_name:
                opt_dict["weight_decay"]= self.weight_decay
                opt_dict["momentum"]= self.momentum

            opt_dict["learning_rate"]= self.learning_rate
            opt_dict["stop_lr"] = self.stop_lr
            opt_dict["stop_step"] = self.stop_step
            opt_dict["decay_step"] = self.decay_step

            opt_dict["train_energy"] = self.train_energy
            opt_dict["train_force"] = self.train_force
            opt_dict["train_ei"] = self.train_ei
            opt_dict["train_virial"] = self.train_virial
            opt_dict["train_egroup"] = self.train_egroup

            opt_dict["start_pre_fac_force"] = self.start_pre_fac_force
            opt_dict["start_pre_fac_etot"] = self.start_pre_fac_etot
            opt_dict["start_pre_fac_ei"] = self.start_pre_fac_ei
            opt_dict["start_pre_fac_virial"] = self.start_pre_fac_virial
            opt_dict["start_pre_fac_egroup"] = self.start_pre_fac_egroup

            opt_dict["end_pre_fac_force"] = self.end_pre_fac_force
            opt_dict["end_pre_fac_etot"] = self.end_pre_fac_etot
            opt_dict["end_pre_fac_ei"] = self.end_pre_fac_ei
            opt_dict["end_pre_fac_virial"] = self.end_pre_fac_virial
            opt_dict["end_pre_fac_egroup"] = self.end_pre_fac_egroup
        elif "SNES" in self.opt_name:
            opt_dict["train_energy"] = self.train_energy
            opt_dict["train_force"] = self.train_force
            opt_dict["train_ei"] = self.train_ei
            opt_dict["train_virial"] = self.train_virial
            opt_dict["train_egroup"] = self.train_egroup
    
            opt_dict["pre_fac_force"] = self.pre_fac_force
            opt_dict["pre_fac_etot"] = self.pre_fac_etot
            opt_dict["pre_fac_ei"] = self.pre_fac_ei
            opt_dict["pre_fac_virial"] = self.pre_fac_virial
            opt_dict["pre_fac_egroup"] = self.pre_fac_egroup

            opt_dict["lambda_1"] =  self.lambda_1
            opt_dict["lambda_2"] =  self.lambda_2
            opt_dict["force_delta"] =  self.force_delta
            opt_dict["population"] =  self.population
            opt_dict["generation"] =  self.generation
        return opt_dict

    def snes_to_nep_txt(self):
        content = ""
        content += "lambda_e    {}\n".format(self.pre_fac_etot)
        content += "lambda_f    {}\n".format(self.pre_fac_force)
        content += "lambda_v    {}\n".format(self.pre_fac_virial)
        content += "batch       {}\n".format(self.batch_size)        
        # content += "lambda_eg   {}\n".format(self.pre_fac_egroup)
        # content += "lambda_ei   {}\n".format(self.pre_fac_ei)
        if self.lambda_1 is not None:
            content += "lambda_1    {}\n".format(self.lambda_1)
        else:
            content += "lambda_1    {}\n".format(-1)
        if self.lambda_2 is not None:
            content += "lambda_2    {}\n".format(self.lambda_2)
        else:
            content += "lambda_2    {}\n".format(-1)
        if self.force_delta is not None:
            content += "force_delta {}\n".format(self.force_delta)
        else:
            content += "force_delta {}\n".format(0)
        if self.population is not None:
            content += "population  {}\n".format(self.population)
        else:
            content += "population  {}\n".format(100)
        if self.generation is not None:
            content += "generation  {}\n".format(self.generation)
        else:
            content += "generation  {}\n".format(10000)
        return content


"""
    dir of explore:
    iter.0000/explore/md:
    --------------------/*-md.job file
    --------------------/*-md.tag.success file
    --------------------/md.000.sys.000/ dir
    -----------------------------------/md.000.sys.000.t.000.p.000 or md.000.sys.000.t.000 dir
    -------------------------------------------------------------/md files: lmp.config, in.lammps, forcefile files, model_devi file
    -------------------------------------------------------------/trajs
    iter.0000/explore/select:
    -------------------/summary.txt
    -------------------/accurate.txt candidate.txt failed.txt candidate_del.txt
    the content of candidate.txt is:
            traj_file_path1 index1
            traj_file_path2 index2
            ...
"""
from pwact.active_learning.slurm.slurm import Mission, SlurmJob, scancle_job
from pwact.utils.slurm_script import get_slurm_job_run_info, split_job_for_group, set_slurm_script_content, recheck_slurm_by_jobtag
from pwact.active_learning.explore.select_image import select_image
from pwact.active_learning.user_input.resource import Resource
from pwact.active_learning.user_input.iter_input import InputParam, MdDetail
from pwact.utils.constant import AL_STRUCTURE, TEMP_STRUCTURE, EXPLORE_FILE_STRUCTURE, TRAIN_FILE_STRUCTUR, \
        FORCEFILED, ENSEMBLE, LAMMPS, LAMMPS_CMD, UNCERTAINTY, DFT_STYLE, SLURM_OUT, SLURM_JOB_TYPE, PWDATA, MODEL_TYPE

from pwact.utils.format_input_output import get_iter_from_iter_name, get_sub_md_sys_template_name,\
    make_md_sys_name, get_md_sys_template_name, make_temp_press_name, make_temp_name, make_lmps_name, make_train_name
from pwact.utils.file_operation import write_to_file, add_postfix_dir, link_file, read_data, search_files, copy_dir, copy_file, del_file, del_dir, del_file_list, del_file_list_by_patten, mv_file
from pwact.utils.draw.hist_model_devi import draw_hist_list
from pwact.utils.app_lib.lammps import make_lammps_input, make_lammps_input_from_lmp_in_file
from pwact.data_format.configop import save_config, get_atom_type

from pwdata import Config
import os
import glob
import numpy as np
import pandas as pd
"""
md_dir:
  a. pwmat+dpkf run md ->MOVEMENT
md_dpkf_dir:
  b. step a.MOVEMENT add Atomic-Energy block ->MOVEMENT
md_traj_dir:
  c. step a.MOVEMENT add Atomic-Energy block ->seperate MOVEMENT to atom.configs
kpu_dir:
  d. step b. calculate KPU of each image(MD)
  f. step d. select cadidate set by limited Delta0 and Delta1
"""
class Explore(object):
    @staticmethod
    def kill_job(root_dir:str, itername:str):
        explore_dir = os.path.join(root_dir, itername, TEMP_STRUCTURE.tmp_run_iter_dir, AL_STRUCTURE.explore)
        md_dir = os.path.join(explore_dir, EXPLORE_FILE_STRUCTURE.md)
        scancle_job(md_dir)

    def __init__(self, itername:str, resource: Resource, input_param:InputParam):
        self.itername = itername
        self.iter = get_iter_from_iter_name(self.itername)
        self.resource = resource
        self.input_param = input_param
        # self.sys_paths = self.input_param.explore.sys_configs
        self.md_job = self.input_param.explore.md_job_list[self.iter]
        # train work dir
        self.train_dir = os.path.join(self.input_param.root_dir, self.itername, TEMP_STRUCTURE.tmp_run_iter_dir, AL_STRUCTURE.train)
        self.real_train_dir = os.path.join(self.input_param.root_dir, self.itername, AL_STRUCTURE.train)
        
        # md work dir
        self.explore_dir = os.path.join(self.input_param.root_dir, itername, TEMP_STRUCTURE.tmp_run_iter_dir, AL_STRUCTURE.explore)
        self.md_dir = os.path.join(self.explore_dir, EXPLORE_FILE_STRUCTURE.md)
        self.select_dir = os.path.join(self.explore_dir, EXPLORE_FILE_STRUCTURE.select)
        self.direct_dir = os.path.join(self.explore_dir, EXPLORE_FILE_STRUCTURE.direct)
        self.kpu_dir = os.path.join(self.explore_dir, EXPLORE_FILE_STRUCTURE.kpu) # for kpu calculate
       
        self.real_explore_dir = os.path.join(self.input_param.root_dir, itername, AL_STRUCTURE.explore)
        self.real_md_dir = os.path.join(self.real_explore_dir, EXPLORE_FILE_STRUCTURE.md)
        self.real_select_dir = os.path.join(self.real_explore_dir, EXPLORE_FILE_STRUCTURE.select)
        self.real_direct_dir = os.path.join(self.real_explore_dir, EXPLORE_FILE_STRUCTURE.direct)
        self.real_kpu_dir = os.path.join(self.real_explore_dir, EXPLORE_FILE_STRUCTURE.kpu) # for kpu calculate

    def back_explore(self):
        slurm_remain, slurm_success = get_slurm_job_run_info(self.real_md_dir, \
        job_patten="*-{}".format(EXPLORE_FILE_STRUCTURE.md_job), \
        tag_patten="*-{}".format(EXPLORE_FILE_STRUCTURE.md_tag),
        for_back = True)
        slurm_done = True if len(slurm_remain) == 0 and len(slurm_success) > 0 else False
        if slurm_done:
            # bk and do new job
            target_bk_file = add_postfix_dir(self.real_explore_dir, postfix_str="bk")
            mv_file(self.real_explore_dir, target_bk_file)
            # if the temp_work_dir/explore exists, delete the train dir
            if os.path.exists(self.explore_dir):
                del_dir(self.explore_dir)

    def make_md_work(self):
        md_work_list = []
        if "lmp_mpi" in self.resource.explore_resource.command.lower():
            is_lmps_26_3 = False
        else:
            is_lmps_26_3 = True
        if self.input_param.train.model_type == MODEL_TYPE.dp:
            matpl = "matpl/dp" if is_lmps_26_3 else "matpl"
            kokkos_head=False
        else:
            if "kokkos" in self.resource.explore_resource.command.lower():
                kokkos_head = True
                matpl = "matpl/nep/kk"
            else:
                kokkos_head = False
                matpl = "matpl/nep"  if is_lmps_26_3 else "matpl"

        for md_index, md in enumerate(self.md_job):
            for sys_id, sys_index in enumerate(md.sys_idx):
                char_len = 3 if len(md.sys_idx) < 1000 else len(str(len(md.sys_idx)))
                md_sys_name = make_md_sys_name(md_index, sys_index, char_len)
                md_sys_dir = os.path.join(self.md_dir, md_sys_name)
                if not os.path.exists(md_sys_dir):
                    os.makedirs(md_sys_dir)
                if md.use_lmps_in:
                    # for lmp_id, lmps_idx in enumerate(md.lmp_in_idx):
                    temp_name = make_lmps_name(md_index, sys_index, sys_id, char_len)
                    temp_dir = os.path.join(md_sys_dir, temp_name)
                    # mkdir: md.000.sys.000/md.000.sys.000.lmps.000
                    if not os.path.exists(temp_dir):
                            os.makedirs(temp_dir)
                    self.set_md_files(len(md_work_list), temp_dir, sys_index, sys_id, None, md, kokkos_head=kokkos_head, matpl=matpl)
                    md_work_list.append(temp_dir)
                else:
                    for temp_index, temp in enumerate(md.temp_list):
                        if ENSEMBLE.nvt in md.ensemble:#for nvt ensemble
                            temp_name = make_temp_name(md_index, sys_index, temp_index, char_len)
                            temp_dir = os.path.join(md_sys_dir, temp_name)
                            # mkdir: md.000.sys.000/md.000.sys.000.t.000
                            if not os.path.exists(temp_dir):
                                os.makedirs(temp_dir)
                            self.set_md_files(len(md_work_list), temp_dir, sys_index, temp_index, None, md, kokkos_head=kokkos_head, matpl=matpl)
                            md_work_list.append(temp_dir)
                        elif ENSEMBLE.npt in md.ensemble: # for npt ensemble
                            for press_index, press in enumerate(md.press_list):
                                temp_press_name = make_temp_press_name(md_index, sys_index, temp_index, press_index, char_len)
                                temp_press_dir = os.path.join(md_sys_dir, temp_press_name)
                                # mkdir: md.000.sys.000/md.000.sys.000.p.000.t.000
                                if not os.path.exists(temp_press_dir):
                                    os.makedirs(temp_press_dir)
                                self.set_md_files(len(md_work_list), temp_press_dir, sys_index, temp_index, press_index, md, kokkos_head=kokkos_head, matpl=matpl)
                                md_work_list.append(temp_press_dir)
                           
        self.make_md_slurm_jobs(md_work_list)
             
    def make_md_slurm_jobs(self, md_work_list:list[str]):
        # delete old job file
        del_file_list_by_patten(self.md_dir, "*{}".format(EXPLORE_FILE_STRUCTURE.md_job))
        group_list = split_job_for_group(self.resource.explore_resource.group_size, md_work_list, self.resource.explore_resource.parallel_num)
        for g_index, group in enumerate(group_list):
            if group[0] == "NONE":
                continue
            jobname = "md{}".format(g_index)
            tag_name = "{}-{}".format(g_index, EXPLORE_FILE_STRUCTURE.md_tag)
            tag = os.path.join(self.md_dir, tag_name)

            # if self.resource.explore_resource.gpu_per_node > 0:
            #     if self.input_param.strategy.uncertainty.upper() == UNCERTAINTY.committee.upper():
            #         gpu_per_node = 1
            #         cpu_per_node = 1
            #         run_cmd = "mpirun -np {} {} -in {} > {} ".format(1, LAMMPS_CMD.lmp_mpi_gpu, LAMMPS.input_lammps, SLURM_OUT.md_out)
            #     else:
            #         cpu_per_node = self.resource.explore_resource.gpu_per_node
            #         gpu_per_node = self.resource.explore_resource.gpu_per_node
            #         run_cmd = "mpirun -np {} {} -in {} > {} ".format(self.resource.explore_resource.gpu_per_node, LAMMPS_CMD.lmp_mpi_gpu, LAMMPS.input_lammps, SLURM_OUT.md_out)
            # else:
            #     cpu_per_node = self.resource.explore_resource.cpu_per_node
            #     run_cmd = "mpirun -np {} {} -in {} > {} ".format(self.resource.explore_resource.cpu_per_node, LAMMPS_CMD.lmp_mpi, LAMMPS.input_lammps, SLURM_OUT.md_out)

            run_cmd = self.resource.explore_resource.command

            group_slurm_script = set_slurm_script_content(
                            gpu_per_node=self.resource.explore_resource.gpu_per_node, 
                            number_node = self.resource.explore_resource.number_node, #1
                            cpu_per_node = self.resource.explore_resource.cpu_per_node,
                            queue_name = self.resource.explore_resource.queue_name,
                            custom_flags = self.resource.explore_resource.custom_flags,
                            env_script = self.resource.explore_resource.env_script,
                            job_name = jobname,
                            run_cmd_template = run_cmd,
                            group = group,
                            job_tag = tag,
                            task_tag = EXPLORE_FILE_STRUCTURE.md_tag,
                            task_tag_faild = EXPLORE_FILE_STRUCTURE.md_tag_faild,
                            parallel_num=self.resource.explore_resource.parallel_num,
                            check_type=None
                            )
            slurm_script_name = "{}-{}".format(g_index, EXPLORE_FILE_STRUCTURE.md_job)
            slurm_job_file = os.path.join(self.md_dir, slurm_script_name)
            write_to_file(slurm_job_file, group_slurm_script, "w")

    '''
    description: 
        waiting: if need set group size, make new script: work1 wait; work2 wait; ...
    param {*} self
    param {list} md_work_list
    return {*}
    author: wuxingxing
    '''    
    def do_md_jobs(self):
        mission = Mission()
        slurm_remain, slurm_success = get_slurm_job_run_info(self.md_dir, \
            job_patten="*-{}".format(EXPLORE_FILE_STRUCTURE.md_job), \
            tag_patten="*-{}".format(EXPLORE_FILE_STRUCTURE.md_tag))
        # for slurm remain, check if tags done
        if len(slurm_remain) > 0:
            #recover slurm jobs
            if len(slurm_remain) > 0:
                print("Run these MD Jobs:\n")
                print(slurm_remain)
                for i, script_path in enumerate(slurm_remain):
                    slurm_job = SlurmJob(lmps_tolerance = self.input_param.strategy.lmps_tolerance)
                    tag_name = "{}-{}".format(os.path.basename(script_path).split('-')[0].strip(), EXPLORE_FILE_STRUCTURE.md_tag)
                    tag = os.path.join(os.path.dirname(script_path),tag_name)
                    slurm_job.set_tag(tag, job_type=SLURM_JOB_TYPE.lammps)
                    slurm_job.set_cmd(script_path)
                    mission.add_job(slurm_job)

            if len(mission.job_list) > 0:
                mission.commit_jobs()
                mission.check_running_job()
                mission.all_job_finished(error_type=SLURM_OUT.md_out)
                
                        
    def set_md_files(self, md_index:int, md_dir:str, sys_index:int, temp_index:int, press_index:int, md_detail:MdDetail, kokkos_head=False, matpl="matpl/nep"):
        target_config = save_config(config=md_detail.config_file_list[sys_index],
                                    input_format=md_detail.config_file_format[sys_index],
                                    wrap = False, 
                                    direct = True, 
                                    sort = False, 
                                    save_format=PWDATA.lammps_lmp, 
                                    save_path=md_dir, 
                                    save_name=LAMMPS.lammps_sys_config)
        # import dpdata
        # _config = dpdata.System(md_detail.config_file_list[sys_index], fmt=md_detail.config_file_format[sys_index])
        # target_config = os.path.join(md_dir, LAMMPS.lammps_sys_config)
        # _config.to("lammps/lmp", target_config, frame_idx=0)
        #2. set forcefiled file
        md_model_paths = self.set_forcefiled_file(md_dir)
        
        #3. set lammps input file
        input_lammps_file = os.path.join(md_dir, LAMMPS.input_lammps)
        # get atom type
        atom_type_list, atomic_number_list = get_atom_type(md_detail.config_file_list[sys_index], md_detail.config_file_format[sys_index])
        atom_type_file = os.path.join(md_dir, LAMMPS.atom_type_file)
        write_to_file(atom_type_file, " ".join(atom_type_list), "w")
        restart_file = search_files(md_dir, "lmps.restart.*")
        restart = 1 if len(restart_file) > 0 else 0 
        if md_detail.use_lmps_in: # lammps.in file from user input
            lmp_input_content = make_lammps_input_from_lmp_in_file(
                            md_file=LAMMPS.lammps_sys_config, #save_file
                            md_type = self.input_param.strategy.md_type,
                            forcefiled = md_model_paths,
                            lmp_in_file = md_detail.lmp_in_file_list[sys_index],
                            atom_type = atomic_number_list,
                            trj_freq = md_detail.trj_freq,
                            boundary=True, #true is 'p p p', false is 'f f f'
                            merge_traj=md_detail.merge_traj,
                            restart = restart,
                            model_deviation_file = EXPLORE_FILE_STRUCTURE.model_devi,
                            matpl = matpl
            )
        
        else: # lammps.in from param.json
            press=md_detail.press_list[press_index] if press_index is not None else None
            lmp_input_content = make_lammps_input(
                            md_file=LAMMPS.lammps_sys_config, #save_file
                            md_type = self.input_param.strategy.md_type,
                            forcefiled = md_model_paths,
                            atom_type = atomic_number_list,
                            ensemble = md_detail.ensemble,
                            nsteps = md_detail.nsteps,
                            dt = md_detail.md_dt,
                            neigh_modify = md_detail.neigh_modify,
                            trj_freq = md_detail.trj_freq,
                            mass = md_detail.mass,
                            temp = md_detail.temp_list[temp_index],
                            tau_t=md_detail.taut, # for fix
                            press=press,
                            tau_p=md_detail.taup if press is not None else None, # for fix    
                            boundary=True, #true is 'p p p', false is 'f f f'
                            merge_traj=md_detail.merge_traj,
                            restart = restart,
                            model_deviation_file = EXPLORE_FILE_STRUCTURE.model_devi,
                            kokkos_head=kokkos_head,
                            matpl = matpl
            )
        
        write_to_file(input_lammps_file, lmp_input_content, "w")
        if md_detail.merge_traj is False:
            traj_dir = os.path.join(md_dir, "traj")
            if not os.path.exists(traj_dir):
                os.makedirs(traj_dir)
        
    '''
    description: 
    param {*} self
    param {str} md_dir
    param {list} forcefile
    return {*}
    author: wuxingxing
    '''    
    def set_forcefiled_file(self, md_dir:str):
        model_name = ""
        md_model_paths = []
        if self.input_param.train.model_type == MODEL_TYPE.nep:
            model_name += "{}/{}".format(TRAIN_FILE_STRUCTUR.model_record, TRAIN_FILE_STRUCTUR.nep_model_lmps)
        elif self.input_param.train.model_type == MODEL_TYPE.dp:
            if self.input_param.strategy.md_type == FORCEFILED.libtorch_lmps:
                model_name += "{}/{}".format(TRAIN_FILE_STRUCTUR.model_record, TRAIN_FILE_STRUCTUR.script_dp_name)
            elif self.input_param.strategy.md_type == FORCEFILED.fortran_lmps:
                if self.input_param.strategy.compress:
                    raise Exception("ERROR! The compress model does not support fortran lammps md! Please change the 'md_type' to 2!")
                else:
                    model_name += "{}/{}".format(TRAIN_FILE_STRUCTUR.fortran_dp, TRAIN_FILE_STRUCTUR.fortran_dp_name)

        for model_index in range(self.input_param.strategy.model_num):
            model_name_i = "{}/{}".format(make_train_name(model_index), model_name)
            source_model_path = os.path.join(self.real_train_dir, model_name_i)
            target_model_path = os.path.join(md_dir, "{}_{}".format(model_index, os.path.basename(source_model_path)))
            link_file(source_model_path, target_model_path)
            md_model_paths.append(target_model_path)
        return md_model_paths
        
    '''
    description: 
        1. copy the explore/md under temp_work_dir to iter*/explore
        2. if reserve traj is false, delete the trajs under iter*/explore/md/*/traj dir
           if use kpu, copy the kpu_model_devi.out to iter*/explore
           delte slurm*.out log files and lammps.log file
        4. copy the explore/select under temp_work_dir to iter*/explore
    param {*} self
    return {*}
    author: wuxingxing
    '''    
    def post_process_md(self):
        md_sys_dir_list = search_files(self.md_dir, get_md_sys_template_name())
        # draw model deviation
        is_kpu = self.input_param.strategy.uncertainty.upper() == UNCERTAINTY.kpu.upper()
        if is_kpu is False:
            for md_sys_dir in md_sys_dir_list:
                draw_hist_list(file_path=[md_sys_dir], 
                                legend_label=["{}-{}".format(self.itername, os.path.basename(md_sys_dir))], 
                                save_path = os.path.join(md_sys_dir, "model_devi_distribution.png"),
                                low=self.input_param.strategy.lower_model_devi_f, 
                                high=self.input_param.strategy.upper_model_devi_f)

        for md_sys_dir in md_sys_dir_list:
            sub_md_sys_dir_list =search_files(md_sys_dir, get_md_sys_template_name())
            for sub_md_sys in sub_md_sys_dir_list:
                target_dir = sub_md_sys.replace(TEMP_STRUCTURE.tmp_run_iter_dir, "")
                copy_dir(sub_md_sys, target_dir)# delete trajs
                if self.input_param.reserve_md_traj and self.input_param.reserve_work:
                    pass
                else:
                    del_file_list([os.path.join(target_dir, EXPLORE_FILE_STRUCTURE.traj)])
                if is_kpu:
                    kpu_source_file = os.path.join(self.kpu_dir, os.path.basename(md_sys_dir), os.path.basename(sub_md_sys), EXPLORE_FILE_STRUCTURE.kpu_model_devi)
                    kpu_taget_file = os.path.join(target_dir, EXPLORE_FILE_STRUCTURE.kpu_model_devi)
                    if os.path.exists(kpu_source_file):
                        copy_file(kpu_source_file, kpu_taget_file)
                if self.input_param.reserve_work is False:#delete lammps.log
                    del_file(os.path.join(target_dir, LAMMPS.log_lammps))
                md_slurms = search_files(self.real_md_dir, "slurm-*") #delete slurm log files
                del_file_list(md_slurms)
                md_tag_slurm_scripts = search_files(self.md_dir, "*md.success")
                md_slurm_scripts = search_files(self.md_dir, "*md.job")
                for file in md_tag_slurm_scripts:
                    copy_file(file, os.path.join(self.real_md_dir, os.path.basename(file)))
                for file in md_slurm_scripts:
                    copy_file(file, os.path.join(self.real_md_dir, os.path.basename(file)))
            if os.path.exists(os.path.join(md_sys_dir, "model_devi_distribution.png")):
                copy_file(os.path.join(md_sys_dir, "model_devi_distribution.png"),
                    os.path.join(self.real_md_dir, os.path.basename(md_sys_dir), "model_devi_distribution.png"))
                link_file(
                    os.path.join(self.real_md_dir, os.path.basename(md_sys_dir), "model_devi_distribution.png"),
                    os.path.join(self.select_dir, "model_devi_distribution-{}.png".format(os.path.basename(md_sys_dir)))
                )
        copy_dir(self.select_dir, self.real_select_dir)
        if self.input_param.strategy.direct:
            # delete nouse files
            del_file_list_by_patten(self.direct_dir, "slurm-*")
            del_file_list_by_patten(self.direct_dir, "*tag*")
            # copy direct dir
            copy_dir(source_dir=self.direct_dir,
                    target_dir=self.real_direct_dir)
            
    
    '''
    description: 
    select structure of system by model deviation (committee method)
    the candidate.csv columns is ["devi_force", "file_path", "config_index"]
    param {*} self
    return {*}
    author: wuxingxing
    '''    
    def select_image_by_committee(self):
        #1. get model_deviation file
        model_deviation_patten = "{}/{}".format(get_sub_md_sys_template_name(), EXPLORE_FILE_STRUCTURE.model_devi)
        model_devi_files = search_files(self.md_dir, model_deviation_patten)
        model_devi_files = sorted(model_devi_files)
        
        #2. for each file, read the model_deviation
        devi_pd = pd.DataFrame(columns=EXPLORE_FILE_STRUCTURE.devi_columns)
        for devi_file in model_devi_files:
            devi_force = read_data(devi_file, skiprows=0)
            tmp_pd = pd.DataFrame()
            tmp_pd[EXPLORE_FILE_STRUCTURE.devi_columns[0]] = devi_force[:, 1]
            tmp_pd[EXPLORE_FILE_STRUCTURE.devi_columns[1]] = devi_force[:, 0]
            tmp_pd[EXPLORE_FILE_STRUCTURE.devi_columns[2]] = os.path.dirname(devi_file)
            devi_pd = pd.concat([devi_pd, tmp_pd])
        devi_pd.reset_index(drop=True, inplace=True)
        devi_pd["config_index"].astype(int)
        #3. select images with lower and upper limitation
        summary_info, summary = select_image(save_dir=self.select_dir, 
                        devi_pd=devi_pd, 
                        lower=self.input_param.strategy.lower_model_devi_f, 
                        higer=self.input_param.strategy.upper_model_devi_f, 
                        max_select=self.input_param.strategy.max_select)
        print("Image select result:\n {}\n\n".format(summary_info))
        return summary

    
    def make_drct_work(self):
        md_work_list = []
        # read candidate.csv
        candidate = pd.read_csv(os.path.join(self.select_dir, EXPLORE_FILE_STRUCTURE.candidate))
        # make scf work dir
        image_list = None
        for index, row in candidate.iterrows():
            config_index    = int(row["config_index"])
            sub_md_sys_path = row["file_path"]
            atom_names = None
            with open(os.path.join(sub_md_sys_path, LAMMPS.atom_type_file), 'r') as rf:
                line = rf.readline()
                atom_names = line.split()
            if image_list is None:
                image_list = Config(data_path=os.path.join(sub_md_sys_path, EXPLORE_FILE_STRUCTURE.traj, "{}{}".format(config_index, LAMMPS.traj_postfix)), 
                                    format=PWDATA.lammps_dump, atom_names=atom_names)
            else:
                image_list.append(Config(data_path=os.path.join(sub_md_sys_path, EXPLORE_FILE_STRUCTURE.traj, "{}{}".format(config_index, LAMMPS.traj_postfix)), 
                                    format=PWDATA.lammps_dump, atom_names=atom_names))
        # cvt_lammps.dump to extxyz
        image_list.to(data_path=self.direct_dir, format=PWDATA.extxyz, data_name="{}".format(EXPLORE_FILE_STRUCTURE.candidate_xyz))
        # copy direct script
        copy_file(self.input_param.strategy.direct_script, os.path.join(self.direct_dir, os.path.basename(self.input_param.strategy.direct_script)))
        # make direct work slurm file
                           
        self.make_drct_slurm_jobs([self.direct_dir])

    def make_drct_slurm_jobs(self, md_work_list:list[str]):
        # delete old job file
        del_file_list_by_patten(self.direct_dir, "*{}".format(EXPLORE_FILE_STRUCTURE.direct_job))
        group_list = split_job_for_group(1, md_work_list, 1)
        for g_index, group in enumerate(group_list):
            if group[0] == "NONE":
                continue
            jobname = "direct{}".format(g_index)
            tag_name = "{}-{}".format(g_index, EXPLORE_FILE_STRUCTURE.direct_tag)
            tag = os.path.join(self.direct_dir, tag_name)

            run_cmd = self.resource.direct_resource.command

            group_slurm_script = set_slurm_script_content(
                            gpu_per_node=self.resource.direct_resource.gpu_per_node, 
                            number_node = self.resource.direct_resource.number_node, #1
                            cpu_per_node = self.resource.direct_resource.cpu_per_node,
                            queue_name = self.resource.direct_resource.queue_name,
                            custom_flags = self.resource.direct_resource.custom_flags,
                            env_script = self.resource.direct_resource.env_script,
                            job_name = jobname,
                            run_cmd_template = run_cmd,
                            group = group,
                            job_tag = tag,
                            task_tag = EXPLORE_FILE_STRUCTURE.direct_tag,
                            task_tag_faild = EXPLORE_FILE_STRUCTURE.direct_tag_faild,
                            parallel_num=self.resource.direct_resource.parallel_num,
                            check_type=None
                            )
            slurm_script_name = "{}-{}".format(g_index, EXPLORE_FILE_STRUCTURE.direct_job)
            slurm_job_file = os.path.join(self.direct_dir, slurm_script_name)
            write_to_file(slurm_job_file, group_slurm_script, "w")

    def do_drct_jobs(self):
        mission = Mission()
        slurm_remain, slurm_success = get_slurm_job_run_info(self.direct_dir, \
            job_patten="*-{}".format(EXPLORE_FILE_STRUCTURE.direct_job), \
            tag_patten="*-{}".format(EXPLORE_FILE_STRUCTURE.direct_tag))
        # for slurm remain, check if tags done
        slurm_done = True if len(slurm_remain) == 0 and len(slurm_success) > 0 else False
        if slurm_done is False:
            slurm_remain = recheck_slurm_by_jobtag(slurm_remain, EXPLORE_FILE_STRUCTURE.md_tag)
        if len(slurm_remain) > 0:
            #recover slurm jobs
            if len(slurm_remain) > 0:
                print("Run these MD Direct Jobs:\n")
                print(slurm_remain)
                for i, script_path in enumerate(slurm_remain):
                    slurm_job = SlurmJob()
                    tag_name = "{}-{}".format(os.path.basename(script_path).split('-')[0].strip(), EXPLORE_FILE_STRUCTURE.direct_tag)
                    tag = os.path.join(os.path.dirname(script_path),tag_name)
                    slurm_job.set_tag(tag, job_type=SLURM_JOB_TYPE.direct)
                    slurm_job.set_cmd(script_path)
                    mission.add_job(slurm_job)

            if len(mission.job_list) > 0:
                mission.commit_jobs()
                mission.check_running_job()
                mission.all_job_finished()
    
    def post_drct(self):
        select_idx = np.loadtxt(os.path.join(self.direct_dir, "{}".format(EXPLORE_FILE_STRUCTURE.select_idx)))
        count = 0
        with open(os.path.join(self.direct_dir, EXPLORE_FILE_STRUCTURE.candidate_xyz), 'r') as file:
            for line in file:
                count += line.count("Lattice")
        
        summary = "Direct selection: Before : {} configs, After : {} selected.\n\n".format(select_idx.shape[0], count)
        write_to_file(os.path.join(self.input_param.root_dir, EXPLORE_FILE_STRUCTURE.iter_select_file), summary, mode='a')
        print(summary)
        return summary


from pwact.utils.json_operation import get_parameter, get_required_parameter

class NetParam(object):
    def __init__(self, net_type:str) -> None:
        self.net_type = net_type

    def set_params(self, network_size: list, bias:bool, resnet_dt:bool, activation:str, physical_property:list=None):
        self.physical_property = physical_property
        self.network_size = network_size
        self.bias = bias
        self.resnet_dt = resnet_dt
        self.activation=activation      
        
    def to_dict(self):
        dicts = {}
        if self.network_size is not None:
            dicts["network_size"] = self.network_size
        else:
            dicts["network_size"] = []
        if "type_" in self.net_type:
            dicts["physical_property"] = self.physical_property
        dicts["bias"] = self.bias
        dicts["resnet_dt"] = self. resnet_dt 
        dicts["activation"] = self.activation
        return dicts
    
    def to_dict_std(self):
        dicts = {}
        if self.network_size is not None:
            dicts["network_size"] = self.network_size
        if "type_" in self.net_type:
            dicts["physical_property"] = self.physical_property
        #dicts["bias"] = self.bias, 
        # if self.resnet_dt is False:
        #     dicts["resnet_dt"] = self.resnet_dt
        #dicts["activation"] = self.activation
        return dicts

class ModelParam(object):
    def __init__(self) -> None:
        self.type_embedding_net = None
        self.embedding_net = None
        self.fitting_net = None
        self.nep_param:NetParam = None

    '''
    description: 
        if type_embedding in first layer of json file is True:
           if type_embedding is not set under layer of model/descriptor, then, the default type embedding set will be used
        else:
           if type_embedding is set under layer of model/descriptor, then the type embedding set will be used
    param {*} self
    param {bool} type_embedding
    param {dict} descriptor_dict
    return {*}
    author: wuxingxing
    '''
    def set_type_embedding_net(self, network_size:list, bias:bool, resnet_dt:bool, activation:str, physical_property:list):
        self.type_embedding_net = NetParam("type_embedding_net")
        self.type_embedding_net.set_params(network_size, bias, resnet_dt, activation, physical_property)
    
    def set_embedding_net(self, network_size:list, bias:bool, resnet_dt:bool, activation:str):
        self.embedding_net = NetParam("embedding_net")
        self.embedding_net.set_params(network_size, bias, resnet_dt, activation)

    def set_dp_fitting_net(self, fitting_net_dict:dict):
        # fitting_net_dict = get_parameter("fitting_net",json_input, {})
        network_size = get_parameter("network_size", fitting_net_dict, [50, 50, 50, 1])
        if network_size[-1] != 1:
            raise Exception("Error: The last layer of the fitting network should have a size of 1 for etot energy, but the input size is {}!".format(network_size[-1]))
        bias = get_parameter("bias", fitting_net_dict, True)
        resnet_dt = get_parameter("resnet_dt", fitting_net_dict, True)
        activation = get_parameter("activation", fitting_net_dict, "tanh")
        self.fitting_net = NetParam("fitting_net")
        self.fitting_net.set_params(network_size, bias, resnet_dt, activation)

    def set_nn_fitting_net(self, fitting_net_dict:dict):
        # fitting_net_dict = get_parameter("fitting_net",json_input, {})
        network_size = get_parameter("network_size", fitting_net_dict,[15,15,1])
        if not isinstance(network_size, list):
            network_size = [network_size]
        if len(network_size) > 1 and network_size[-1] != 1:
            raise Exception("Error: The last layer of the fitting network should have a size of 1 for etot energy, but the input size is {}!".format(network_size[-1]))
        bias = True # get_parameter("bias", fitting_net_dict, True)
        resnet_dt = False # get_parameter("resnet_dt", fitting_net_dict, False)
        activation = "tanh" #get_parameter("activation", fitting_net_dict, )
        self.fitting_net = NetParam("fitting_net")
        self.fitting_net.set_params(network_size, bias, resnet_dt, activation)

    # def to_dict(self):
    #     # dicts = {}
    #     # if self.embedding_net is not None:
    #     #     dicts[self.embedding_net.net_type] = self.embedding_net.to_dict()
    #     # if self.fitting_net is not None:
    #     #     dicts[self.fitting_net.net_type] = self.fitting_net.to_dict()
    #     return self.fitting_net.to_dict_std()




from ase.io import read
from fairchem.core import OCPCalculator
import os
output_file = 'train.xyz'
traj = read("select.xyz", index=":")
calc = OCPCalculator(
    checkpoint_path="/share/public/PWMLFF_test_data/eqv2-models/eqV2_31M_omat.pt",
    cpu=False,
)

def atoms2xyzstr(atoms):
    num_atom = atoms.get_global_number_of_atoms()
    vol = atoms.get_volume()
    pos = atoms.positions
    forces = atoms.get_forces()
    energy = atoms.get_potential_energy()
    cell = atoms.cell
    virial = -atoms.get_stress(voigt=False) * vol
    xyzstr = "%d\n" % num_atom
    xyz_head = 'Lattice="%.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f" Properties=species:S:1:pos:R:3:forces:R:3 energy=%.8f' 
    xyz_format = (cell[0,0],cell[0,1],cell[0,2],cell[1,0],cell[1,1],cell[1,2],cell[2,0],cell[2,1],cell[2,2],energy)
    if virial is not None:
        xyz_head += ' virial="%.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f %.8f"'
        xyz_format += (
            virial[0,0], virial[0,1], virial[0,2],
            virial[1,0], virial[1,1], virial[1,2],
            virial[2,0], virial[2,1], virial[2,2]
            )
    xyz_head += '\n'
    xyzstr += xyz_head % xyz_format
    for i in range(num_atom):
        xyzstr += "%2s %14.8f %14.8f %14.8f %14.8f %14.8f %14.8f\n" %\
        (atoms[i].symbol,pos[i,0],pos[i,1],pos[i,2],forces[i,0],forces[i,1],forces[i,2])
    return xyzstr

f = open(output_file, "w")
for i in range(len(traj)):
    atoms = traj[i]
    atoms.calc = calc
    f.write(atoms2xyzstr(atoms))
f.close()




def cmd_infos(cmd_type=None):
    if cmd_type is None:
        cmd_info = ""
        cmd_info += "init_bulk".lower() + "\n"
        cmd_info += "you could use this method to do relax, super cell, scale, pertub and aimd\n"
        cmd_info += "example: pwact init_bulk init_param.json machine.json\n\n"
        
        # cmd_info += "int_surface".lower() + "\n"
        # cmd_info += ""
        cmd_info += "run".lower() + "\n"
        cmd_info += "you could use this method to do active learning sampling\n"
        cmd_info += "example: pwact run param.json machine.json\n\n"
        # cmd_info += "init_json".lower() + "\n"

        # cmd_info += "run_json".lower() + "\n"

        cmd_info += "kill".lower() + "\n"
        cmd_info += "you could use this method to kill active learning tasks currently being executed\n"
        cmd_info += "example:\n"
        cmd_info += "'pwact kill init_bulk' for 'init_bulk' tasks\n"
        cmd_info += "'pwact kill run' for 'run' tasks\n\n"

        cmd_info += "gather_pwdata".lower() + "\n"
        cmd_info += "you could use this method to extract pwdatas after active learing done.\nFor more detail for this command, you could use 'pwact gather_pwdata -h'\n\n"

        cmd_info += "filter" + "\n"
        cmd_info += "you could use this method to test the selection results corresponding to the upper and lower limit settings,  you could use 'pwact filter -h'\n\n"

        cmd_info += "to_pwdata".lower() + "\n"
        cmd_info += "(This method has been abandoned, new conversion methods refer to the pwdata documentation http://doc.lonxun.com/PWMLFF/Appendix-2/)\n\n"
        cmd_info += "you could use this method to change outcars or movements to pwdata format.\nFor more detail for this command, you could use 'pwact to_pwdata -h'\n\n"

        cmd_info += "\n\n"
        cmd_info += "examples of pwact are located at the dir 'pwact/example'\n"
        cmd_info += "you can alse download them from github link:\n"
        cmd_info += "https://github.com/LonxunQuantum/PWact/tree/main/pwact/example\n\n"
        cmd_info += "or download from BaiduNetdisk included the calculation results of examples:\n"
        cmd_info += "https://pan.baidu.com/s/14E0u_7cpntiBZgg-C1S5XA?pwd=pwmt\n\n"

    elif cmd_type == "init_bulk":
        cmd_info = cmd_info_init_bulk()
    elif cmd_type == "run":
        cmd_info = cmd_info_run_iter()
    elif cmd_type == "kill":
        cmd_info = cmd_info_kill()
    elif cmd_type == "filter":
        cmd_info = cmd_info_filter()
    print(cmd_info)


def cmd_info_init_bulk():
    cmd_info = ""
    cmd_info += "init_bulk".lower() + "\n"
    cmd_info += "you could use this method to do relax, super cell, scale, pertub and aimd\n"
    cmd_info += "example: pwact init_bulk init_param.json machine.json\n\n"
    return cmd_info

def cmd_info_run_iter():
    cmd_info = ""
    cmd_info += "run" + "\n"
    cmd_info += "you could use this method to do active learning sampling\n"
    cmd_info += "example: pwact run param.json machine.json\n\n"
    return cmd_info

def cmd_info_kill():
    cmd_info = ""
    cmd_info += "kill".lower() + "\n"
    cmd_info += "you could use this method to kill active learning tasks currently being executed\n"
    cmd_info += "example:\n"
    cmd_info += "'pwact kill init_bulk' for 'init_bulk' tasks\n"
    cmd_info += "'pwact kill run' for 'run' tasks\n\n"
    return cmd_info

def cmd_info_filter():
    cmd_info = ""
    cmd_info += "filter" + "\n"
    cmd_info += "you could use this method to test the selection results corresponding to the upper and lower limit settings\n"
    cmd_info += "example:\n"
    cmd_info += "'pwact filter -i iter.0000/explore/md -l 0.01 -u 0.02 -s'\n\n"
    return cmd_info


"""
this script inculding the functions for data coverting in active learning iters
"""

from pwact.utils.format_input_output import make_iter_name
import os

"""
@Description :
    this function should not be used, when the new labeled data nums are less, just ignore them
    return data paths
@Returns     :
@Author       :wuxingxing
"""
def get_movement_num(iter_result):
    keys = list(iter_result.keys()) #反向
    keys = sorted(keys, key=lambda x:int(x[5:]), reverse=True)

    movement_list = []
    movement_from_iter = []
    for i in keys:
        if "feature_path" not in iter_result[i].keys() and len(iter_result[i]["movement_file"]) > 0:
            movement_list.extend([os.path.join(iter_result[i]["movement_dir"], _) for _ in iter_result[i]["movement_file"]])
            movement_from_iter.append(i)
    
    return movement_list, movement_from_iter

"""
@Description :
get training data path
if first training: return init_data
else: return all feature_path

the init data dir should be set to the last path, when the new scale data info comes,
they will be saved to the init data path.

data_limit: when new datas more than this value, retrain the model.
@Returns     :
@Author       :wuxingxing
"""
def get_feature_data_path(self):
    retain_sign = False
    iter_index = int(self.itername[5:])
    if iter_index == 0:
        return self.system_info["init_data_path"], True

    pre_iter_name = make_iter_name(iter_index-1)
    if "feature_path" not in self.iter_result[pre_iter_name].keys():
        return [], False

    data_paths = set()
    keys = list(self.iter_result.keys()) #反向
    keys = sorted(keys, key=lambda x:int(x[5:]), reverse=True)
    for i in keys:
        if "feature_path" in self.iter_result[i].keys():
            data_paths.add(self.iter_result[i]["feature_path"])
            retain_sign = True
    data_paths = list(data_paths)
    data_paths.extend(self.system_info["init_data_path"])
    # iter_res_path = "{}/iter_result.json".format(system_info["work_root_path"])
    # json.dump(iter_result, open(iter_res_path, "w"), indent=4)
    return data_paths, retain_sign

'''
Description: 
 get dir name of images which under the 'path'
param {*} path
param {*} basename
Returns: 
    ['image_000', 'image_001', ..., 'image_N']
Author: WU Xingxing
'''
def get_image_nums(path, basename="image"):
    dirlist = os.listdir(path)
    res = []
    for d in dirlist:
        if basename in d:
            res.append(d)
    res = sorted(res, key=lambda x: int(x.split('_')[-1]))
    return res



"""
    realx dor strcuture
    --/work_path/
    ------------/relax
    -----------------/*_init_config/*_init_.config, etot.input, pseudo files, final.config, mvms_relax
    ------------/super_cell_scale
    -----------------/*_init_config/scale.config, from final.config or *_init_.config
    ------------/perturb
    --------------------/*_init_config/scale.config or final.config
    ----------------------------------/structures/*.config_perturb
    ------------/AIMD
    -----------------/*_init_config/*_config_dir the config file from pertub or *_init_.config
    -------------------------------/etot.input, atom.config, ...
    -----------/result
    -----------------/*_init_config/relax_mvm_*_init_config if need; aimd_mvm_*_init_config if need
    
"""
import os
import glob
import json
import bisect
from pwact.active_learning.user_input.resource import Resource
from pwact.active_learning.user_input.init_bulk_input import InitBulkParam
from pwact.active_learning.init_bulk.duplicate_scale import get_config_files_with_order

from pwact.utils.constant import PWMAT, INIT_BULK, TEMP_STRUCTURE, SLURM_OUT, DFT_STYLE, PWDATA, VASP
from pwact.active_learning.slurm.slurm import SlurmJob, Mission
from pwact.utils.slurm_script import get_slurm_job_run_info, split_job_for_group, set_slurm_script_content
    
from pwact.utils.file_operation import write_to_file, link_file, del_dir, del_file_list_by_patten, get_random_nums
from pwact.utils.app_lib.common import link_pseudo_by_atom, set_input_script
from pwact.data_format.configop import extract_pwdata, save_config, get_atom_type, load_config

import pandas as pd
from pwdata import Config

# from pwact.utils.constant import DFT_TYPE, VASP, PWDATA, AL_STRUCTURE, TEMP_STRUCTURE,\
#     LABEL_FILE_STRUCTURE, EXPLORE_FILE_STRUCTURE, LAMMPS, SLURM_OUT, DFT_STYLE, PWMAT, INIT_BULK
# from pwact.utils.file_operation import write_to_file, copy_file, copy_dir, search_files, mv_file, add_postfix_dir, del_dir, del_file_list_by_patten, link_file

class Relabel(object):
    def __init__(self, resource: Resource, input_param:InitBulkParam):
        self.resource = resource
        self.input_param = input_param
        self.init_configs = self.input_param.sys_config
        self.relax_dir = os.path.join(self.input_param.root_dir, TEMP_STRUCTURE.tmp_init_bulk_dir, INIT_BULK.relax)
        self.super_cell_scale_dir = os.path.join(self.input_param.root_dir, TEMP_STRUCTURE.tmp_init_bulk_dir, INIT_BULK.super_cell_scale)
        self.pertub_dir = os.path.join(self.input_param.root_dir,TEMP_STRUCTURE.tmp_init_bulk_dir, INIT_BULK.pertub)
        self.aimd_dir = os.path.join(self.input_param.root_dir, TEMP_STRUCTURE.tmp_init_bulk_dir, INIT_BULK.aimd)
        self.real_aimd_dir = os.path.join(self.input_param.root_dir, INIT_BULK.aimd)
        
        self.scf_dir = os.path.join(self.input_param.root_dir, TEMP_STRUCTURE.tmp_init_bulk_dir, INIT_BULK.scf)
        self.real_scf_dir = os.path.join(self.input_param.root_dir, INIT_BULK.scf)

        self.bigmodel_dir = os.path.join(self.input_param.root_dir, TEMP_STRUCTURE.tmp_init_bulk_dir, INIT_BULK.bigmodel)
        self.real_bigmodel_dir = os.path.join(self.input_param.root_dir, INIT_BULK.bigmodel)
        
        self.direct_dir = os.path.join(self.bigmodel_dir, INIT_BULK.direct)
        self.real_direct_dir = os.path.join(self.real_bigmodel_dir, INIT_BULK.direct)

    def check_work_done(self):
        slurm_remain, slurm_success = get_slurm_job_run_info(self.scf_dir, \
            job_patten="*-{}".format(INIT_BULK.scf_job), \
            tag_patten="*-{}".format(INIT_BULK.scf_tag))
        slurm_done = True if len(slurm_remain) == 0 and len(slurm_success) > 0 else False # len(slurm_remain) > 0 exist slurm jobs need to do
        return slurm_done

    def make_scf_work(self):
        def find_position_binary(prefix_sum, N):
            idx = bisect.bisect_right(prefix_sum, N)
            if idx == 0:
                return 0
            elif idx >= len(prefix_sum):
                return (len(prefix_sum)-1)
            else:
                return idx

        def compute_prefix_sum(arr):
            prefix_sum = []
            current_sum = 0
            for num in arr:
                current_sum += num
                prefix_sum.append(current_sum)
            return prefix_sum
                            
        candidate = Config(data_path=os.path.join(self.direct_dir, INIT_BULK.direct_traj), format=PWDATA.extxyz)
        # from idx get config idx
        candidate_idx = json.load(open(os.path.join(self.direct_dir, INIT_BULK.candidate_idx)))
        candidate_idx_sum = compute_prefix_sum([candidate_idx[_]['num'] for _ in candidate_idx.keys()])
        _tmp = Config(data_path=os.path.join(self.direct_dir, INIT_BULK.direct_traj), format=PWDATA.extxyz)
        scf_dir_list = []
        if self.input_param.dft_input.scf_max_num is not None:
            random_list = get_random_nums(0, len(candidate.images), self.input_param.dft_input.scf_max_num, seed=2024)
        else:
            random_list = None
        for index, image in enumerate(candidate.images):
            if random_list is not None and index not in random_list:
                continue
            _idx = find_position_binary(candidate_idx_sum, index)
            config_idx = candidate_idx["{}".format(_idx)]['idx']
            scf_dir = os.path.join(self.scf_dir, "{}".format(index))
            if not os.path.exists(scf_dir):
                os.makedirs(scf_dir)
                
            _tmp.images = [image]
            _tmp.to(data_path=scf_dir, data_name=PWMAT.atom_config,
                         format=PWDATA.pwmat_config)
            self.make_scf_file(
                    scf_dir      =scf_dir, 
                    traj_file    =os.path.join(scf_dir, PWMAT.atom_config), 
                    traj_format  =PWDATA.pwmat_config, 
                    target_format=DFT_STYLE.get_pwdata_format(self.input_param.dft_style, is_cp2k_coord=True),
                    input_file   =self.init_configs[config_idx].scf_input_file,
                    kspacing     =self.init_configs[config_idx].scf_kspacing, 
                    flag_symm    =self.init_configs[config_idx].scf_flag_symm,
                    is_dftb      = False,
                    in_skf       =None)

            scf_dir_list.append(scf_dir)
        
        self.make_scf_slurm_job_files(scf_dir_list)

    def make_scf_slurm_job_files(self, scf_dir_list:list[str]):
        del_file_list_by_patten(self.scf_dir, "*{}".format(INIT_BULK.scf_job))
        group_list = split_job_for_group(self.resource.dft_resource.group_size, scf_dir_list, self.resource.dft_resource.parallel_num)
        for group_index, group in enumerate(group_list):
            if group[0] == "NONE":
                continue
            jobname = "scf{}".format(group_index)
            tag_name = "{}-{}".format(group_index, INIT_BULK.scf_tag)
            tag = os.path.join(self.scf_dir, tag_name)
            run_cmd = self.resource.dft_resource.command
            group_slurm_script = set_slurm_script_content(gpu_per_node=self.resource.dft_resource.gpu_per_node, 
                number_node = self.resource.dft_resource.number_node, 
                cpu_per_node = self.resource.dft_resource.cpu_per_node,
                queue_name = self.resource.dft_resource.queue_name,
                custom_flags = self.resource.dft_resource.custom_flags,
                env_script = self.resource.dft_resource.env_script,
                job_name = jobname,
                run_cmd_template = run_cmd,
                group = group,
                job_tag = tag,
                task_tag = INIT_BULK.scf_tag, 
                task_tag_faild = INIT_BULK.scf_tag_failed,
                parallel_num=self.resource.dft_resource.parallel_num,
                check_type=None
                )
            slurm_script_name = "{}-{}".format(group_index, INIT_BULK.scf_job)
            slurm_job_file =  os.path.join(self.scf_dir, slurm_script_name)
            write_to_file(slurm_job_file, group_slurm_script, "w")

    def do_scf_jobs(self):
        mission = Mission()
        slurm_remain, slurm_success = get_slurm_job_run_info(self.scf_dir, \
            job_patten="*-{}".format(INIT_BULK.scf_job), \
            tag_patten="*-{}".format(INIT_BULK.scf_tag))
        slurm_done = True if len(slurm_remain) == 0 and len(slurm_success) > 0 else False # len(slurm_remain) > 0 exist slurm jobs need to do
        if slurm_done is False:
            #recover slurm jobs
            if len(slurm_remain) > 0:
                print("Run these relabel Jobs:\n")
                print(slurm_remain)
                for i, script_path in enumerate(slurm_remain):
                    slurm_job = SlurmJob()
                    tag_name = "{}-{}".format(os.path.basename(script_path).split('-')[0].strip(), INIT_BULK.scf_tag)
                    tag = os.path.join(os.path.dirname(script_path),tag_name)
                    slurm_job.set_tag(tag)
                    slurm_job.set_cmd(script_path)
                    mission.add_job(slurm_job)

            if len(mission.job_list) > 0:
                mission.commit_jobs()
                mission.check_running_job()
                mission.all_job_finished(error_type=SLURM_OUT.dft_out)
                # mission.move_slurm_log_to_slurm_work_dir()

    def make_scf_file(self, 
                    scf_dir, 
                    traj_file    , 
                    traj_format  , # the input is pwmat/config
                    target_format,
                    input_file   ,
                    kspacing     =None, 
                    flag_symm    =None,
                    is_dftb      =None,
                    in_skf       =None,
                    atom_names:list[str]=None):
        if DFT_STYLE.pwmat == self.resource.dft_style:
            target_config = traj_file
            pass
        else:
            if DFT_STYLE.vasp == self.resource.dft_style: # when do scf, the vasp input file name is 'POSCAR'
                save_name = VASP.poscar
            else:
                save_name="{}".format(DFT_STYLE.get_normal_config(self.resource.dft_style))# for cp2k this param will be set as coord.xzy
            target_config = save_config(config=traj_file,
                                        input_format=traj_format,
                                        wrap = False, 
                                        direct = True, 
                                        sort = True, 
                                        save_name = save_name,
                                        save_format=DFT_STYLE.get_pwdata_format(dft_style=self.resource.dft_style, is_cp2k_coord=True),
                                        save_path=scf_dir, 
                                        atom_names=atom_names)

        #2.
        atomic_name_list, atomic_number_list = get_atom_type(traj_file, traj_format)
        #1. set pseudo files
        pseudo_names = link_pseudo_by_atom(
                pseudo_list = self.input_param.dft_input.pseudo, 
                target_dir = scf_dir, 
                atom_order = atomic_name_list, 
                dft_style = self.resource.dft_style,
                basis_set_file  =self.input_param.dft_input.basis_set_file,
                potential_file  =self.input_param.dft_input.potential_file
                )
    
        #2. make etot.input file
        set_input_script(
            input_file=input_file,
            config=target_config,
            dft_style=self.resource.dft_style,
            kspacing=kspacing, 
            flag_symm=flag_symm, 
            save_dir = scf_dir,
            pseudo_names=pseudo_names,
            gaussian_base_param=self.input_param.dft_input.gaussian_base_param,# these for cp2k
            is_scf = True
        )

    def do_post_process(self):
        if os.path.exists(self.scf_dir):
            link_file(self.scf_dir, self.real_scf_dir)



import numpy as np

default_config = {
    "GLOBAL": {"PROJECT": "DPGEN"},
    "FORCE_EVAL": {
        "METHOD": "QS",
        "STRESS_TENSOR": "ANALYTICAL",
        "DFT": {
            "BASIS_SET_FILE_NAME": "./cp2k_basis_pp_file/BASIS_MOLOPT",
            "POTENTIAL_FILE_NAME": "./cp2k_basis_pp_file/GTH_POTENTIALS",
            "CHARGE": 0,
            "UKS": "F",
            "MULTIPLICITY": 1,
            "MGRID": {"CUTOFF": 400, "REL_CUTOFF": 50, "NGRIDS": 4},
            "QS": {"EPS_DEFAULT": "1.0E-12"},
            "SCF": {"SCF_GUESS": "ATOMIC", "EPS_SCF": "1.0E-6", "MAX_SCF": 50},
            "XC": {"XC_FUNCTIONAL": {"_": "PBE"}},
        },
        "SUBSYS": {
            "CELL": {"A": "10 .0 .0", "B": ".0 10 .0", "C": ".0 .0 10"},
            "COORD": {"@include": "coord.xyz"},
            "KIND": {
                "_": ["H", "C", "N"],
                "POTENTIAL": ["GTH-PBE-q1", "GTH-PBE-q4", "GTH-PBE-q5"],
                "BASIS_SET": ["DZVP-MOLOPT-GTH", "DZVP-MOLOPT-GTH", "DZVP-MOLOPT-GTH"],
            },
        },
        "PRINT": {"FORCES": {"_": "ON"}, "STRESS_TENSOR": {"_": "ON"}},
    },
}


def update_dict(old_d, update_d):
    """A method to recursive update dict
    :old_d: old dictionary
    :update_d: some update value written in dictionary form.
    """
    import collections.abc

    for k, v in update_d.items():
        if (
            k in old_d
            and isinstance(old_d[k], dict)
            and isinstance(update_d[k], collections.abc.Mapping)
        ):
            update_dict(old_d[k], update_d[k])
        else:
            old_d[k] = update_d[k]


def iterdict(d, out_list, flag=None, indent=0):
    """
    :doc: a recursive expansion of dictionary into cp2k input
    :k: current key
    :v: current value
    :d: current dictionary under expansion
    :flag: used to record dictionary state. if flag is None,
    it means we are in top level dict. flag is a string.
    :indent: intent for current section.
    """
    for k, v in d.items():
        k = str(k)  # cast key into string
        # if value is dictionary
        if isinstance(v, dict):
            # flag == None, it is now in top level section of cp2k
            if flag is None:
                out_list.append("&" + k)
                out_list.append("&END " + k)
                iterdict(v, out_list, k, indent + 2)
            # flag is not None, now it has name of section
            else:
                index = out_list.index(" " * (indent - 2) + "&END " + flag)
                out_list.insert(index, " " * indent + "&" + k + " #" + flag)
                out_list.insert(index + 1, " " * indent + "&END " + k + " #" + flag)
                # the flag now contains its parent section name, separed by "#".
                iterdict(v, out_list, k + " #" + flag, indent + 2)
        elif isinstance(v, list):
            #            print("we have encountered the repeat section!")
            index = out_list.index(" " * (indent - 2) + "&" + flag)
            # delete the current constructed repeat section
            del out_list[index : index + 2]
            # do a loop over key and corresponding list
            k_tmp_list = []
            v_list_tmp_list = []
            for k_tmp, v_tmp in d.items():
                k_tmp_list.append(str(k_tmp))
                v_list_tmp_list.append(v_tmp)
            for repeat_keyword in zip(*v_list_tmp_list):
                out_list.insert(index, " " * (indent - 2) + "&" + flag)
                out_list.insert(index + 1, " " * (indent - 2) + "&END " + flag)
                for idx, k_tmp in enumerate(k_tmp_list):
                    if k_tmp == "_":
                        out_list[index] = (
                            " " * (indent - 2)
                            + "&"
                            + flag.split(" #")[0]
                            + " "
                            + repeat_keyword[idx]
                        )
                    else:
                        out_list.insert(
                            index + 1,
                            " " * (indent) + k_tmp + " " + repeat_keyword[idx],
                        )
            break

        else:
            v = str(v)
            if flag is None:
                out_list.append(k + " " + v)
                print(k, ":", v)
            else:
                if k == "_":
                    index = out_list.index(" " * (indent - 2) + "&" + flag)
                    out_list[index] = (
                        " " * (indent - 2) + "&" + flag.split(" #")[0] + " " + v
                    )

                else:
                    index = out_list.index(" " * (indent - 2) + "&END " + flag)
                    out_list.insert(index, " " * indent + k + " " + v)


def make_cp2k_input(sys_data, fp_params):
    # covert cell to cell string
    cell = sys_data["cells"][0]
    cell = np.reshape(cell, [3, 3])
    cell_a = np.array2string(cell[0, :])
    cell_a = cell_a[1:-1]
    cell_b = np.array2string(cell[1, :])
    cell_b = cell_b[1:-1]
    cell_c = np.array2string(cell[2, :])
    cell_c = cell_c[1:-1]

    # get update from user
    user_config = fp_params
    # get update from cell
    cell_config = {
        "FORCE_EVAL": {"SUBSYS": {"CELL": {"A": cell_a, "B": cell_b, "C": cell_c}}}
    }
    update_dict(default_config, user_config)
    update_dict(default_config, cell_config)
    # output list
    input_str = []
    iterdict(default_config, input_str)
    string = "\n".join(input_str)
    return string


def make_cp2k_xyz(sys_data):
    # get structral information
    atom_names = sys_data["atom_names"]
    atom_types = sys_data["atom_types"]

    # write coordinate to xyz file used by cp2k input
    coord_list = sys_data["coords"][0]
    u = np.array(atom_names)
    atom_list = u[atom_types]
    x = "\n"
    for kind, coord in zip(atom_list, coord_list):
        x += str(kind) + " " + str(coord[:])[1:-1] + "\n"
    return x


def make_cp2k_input_from_external(sys_data, exinput_path):
    # read the input content as string
    with open(exinput_path) as f:
        exinput = f.readlines()

    # find the ABC cell string
    for line_idx, line in enumerate(exinput):
        if "ABC" in line:
            delete_cell_idx = line_idx
            delete_cell_line = line

    # remove the useless CELL line
    exinput.remove(delete_cell_line)

    # insert the cell information
    # covert cell to cell string
    cell = sys_data["cells"][0]
    cell = np.reshape(cell, [3, 3])
    cell_a = np.array2string(cell[0, :])
    cell_a = cell_a[1:-1]
    cell_b = np.array2string(cell[1, :])
    cell_b = cell_b[1:-1]
    cell_c = np.array2string(cell[2, :])
    cell_c = cell_c[1:-1]

    exinput.insert(delete_cell_idx, "A  " + cell_a + "\n")
    exinput.insert(delete_cell_idx + 1, "B  " + cell_b + "\n")
    exinput.insert(delete_cell_idx + 2, "C  " + cell_c + "\n")

    return "".join(exinput)



import os
from pwact.utils.constant import DFT_STYLE, CP2K
from pwact.utils.app_lib.pwmat import read_and_check_etot_input
from pwact.utils.json_operation import get_parameter, get_required_parameter
from pwact.utils.app_lib.common import get_vasp_pseudo_atom_type
class SCFParam(object):
    def __init__(self, json_dict:dict, 
        is_relax:bool=False, 
        is_aimd:bool=False, 
        is_scf:bool=False, 
        root_dir:str=None, 
        dft_style:str=None,
        scf_style:str=None,
        is_bigmodel:bool=False,
        is_direct:bool=False) -> None:# for scf relabel in init_bulk
        
        self.dft_style = dft_style
        self.root_dir = root_dir

        self.relax_input_list = []
        self.aimd_input_list = []
        self.scf_input_list = []
        self.pseudo = []
        self.scf_pseudo = []
        self.use_dftb = False

        if is_scf:
            if "scf_input" in json_dict.keys(): # for init_bulk relabel
                if dft_style == DFT_STYLE.bigmodel:
                    self.bigmodel_script = get_required_parameter("bigmodel_script", json_dict)
                else:
                    json_scf = get_required_parameter("scf_input", json_dict)
                    self.scf_input_list = self.set_input(json_scf, flag_symm=0)
            else: # for run_iter
                if dft_style == DFT_STYLE.bigmodel:
                    self.bigmodel_script = get_required_parameter("bigmodel_script", json_dict)
                else:
                    self.scf_input_list = self.set_input(json_dict, flag_symm=0)
                    if self.scf_input_list[0].use_dftb:
                        self.use_dftb = True
        if is_aimd:
            json_aimd = get_required_parameter("aimd_input", json_dict)
            self.aimd_input_list = self.set_input(json_aimd, flag_symm=0)
            if self.aimd_input_list[0].use_dftb:
                self.use_dftb = True
        if is_relax:
            json_relax = get_required_parameter("relax_input", json_dict)
            self.relax_input_list = self.set_input(json_relax, flag_symm=3) 
            if self.relax_input_list[0].use_dftb:
                self.use_dftb = True

        if is_bigmodel: # init_bulk
            json_bigmodel = get_required_parameter("bigmodel_input", json_dict)
            self.bigmodel_input_list = self.set_input(json_bigmodel, flag_symm=3) 

        if is_direct: # init_bulk
            json_direct = get_required_parameter("direct_input", json_dict)
            self.direct_input_list = self.set_input(json_direct, flag_symm=3) 

        self.scf_max_num = get_parameter("scf_max_num", json_dict, None)
        # for pwmat, use 'pseudo' key
        # for vasp is INCAR file, use 'pseudo' key        
        pseudo = get_parameter("pseudo", json_dict, [])
        self.pseudo = self._set_pseudo(pseudo, dft_style)
        
        if is_scf:
            scf_pseudo = get_parameter("scf_pseudo", json_dict, [])
            self.scf_pseudo = self._set_pseudo(scf_pseudo, scf_style)

        # for pwmat-dftb is in_skf, a dir string
        in_skf = get_parameter("in_skf", json_dict, None)
        self.in_skf = None
        if self.use_dftb:
            if in_skf is not None:
                self.in_skf = in_skf if os.path.isabs(in_skf) else os.path.abspath(in_skf)
            #     if not os.path.exists(self.in_skf):
            #         raise Exception("ERROR! The 'in_skf' dir {} not exsit!".format(self.in_skf))
            # else:
            #     raise Exception("ERROR! The 'USE_DFTB' is set in scf.input file, but the 'in_skf' dir not set!")
        # else:
        #     pass
        # for cp2k
        gaussian_param = get_parameter("gaussian_param", json_dict, None)
        if gaussian_param is not None:
            self.basis_set_file = os.path.abspath(get_parameter("basis_set_file", gaussian_param, None))
            self.potential_file = os.path.abspath(get_parameter("potential_file", gaussian_param, None))
            basis_set_list = get_parameter("basis_set_list", gaussian_param, None)
            potential_list = get_parameter("potential_list", gaussian_param, None)
            atom_list = get_parameter("atom_list", gaussian_param, None)
            self.gaussian_base_param = {}
            self.gaussian_base_param["KSPACING"] = get_parameter("kspacing", gaussian_param, None)
            self.gaussian_base_param["ELEMENT"] = atom_list
            self.gaussian_base_param["BASIS_SET"] = basis_set_list
            self.gaussian_base_param["POTENTIAL"] = potential_list
            self.gaussian_base_param["BASIS_SET_FILE_NAME"] = os.path.basename(self.basis_set_file)
            self.gaussian_base_param["POTENTIAL_FILE_NAME"] = os.path.basename(self.potential_file)
        else:
            self.basis_set_file = None# os.path.abspath(get_parameter("basis_set_file", json_dict, None))
            self.potential_file = None#os.path.abspath(get_parameter("potential_file", json_dict, None))
            self.gaussian_base_param = None
            self.kspacing = None
        # for cp2k and pwmat gaussion
        

    def _set_pseudo(self, pseudo, style:str):
        res_pseudo = []
        if isinstance(pseudo, str):
            pseudo = list(pseudo)
        for pf in pseudo:
            if not os.path.exists(pf):
                raise Exception("Error! pseduo file {} does not exist!".format(pf))
            if not os.path.isabs(pf):
                pf = os.path.abspath(pf)
            # for vasp pseudo files, read the pseduo files and get the atom type in pseduo
            if style == DFT_STYLE.vasp:
                atom_type = get_vasp_pseudo_atom_type(pf)
            # for pwmat pseudo files, get the atom type from pseudo file name
            elif style == DFT_STYLE.pwmat:
                atom_type = os.path.basename(pf).split('.')[0]
            res_pseudo.append([pf, atom_type])
        return res_pseudo

    '''
    description: 
        set dft input file
        for PWmat: etot.input
        for Vasp: Incar
        for cp2k: input.inp file

        json_input: could be a string or list string, for input control file or a list files
                    could be a dict ir dict list, mainly for pwmat to set kspacing

    param {*} self
    param {*} json_etot
    param {*} root_dir
    param {int} flag_symm
    return {*}
    author: wuxingxing
    '''    
    def set_input(self, json_input, flag_symm:int):
        input_list = []
        # for vasp incar: the input is a str
        if isinstance(json_input, str):
            input_file = json_input
            if not os.path.isabs(input_file):
                input_file = os.path.join(self.root_dir, input_file)
            if not os.path.exists(input_file):
                raise Exception("Error! The input file {} does not exist!".format(input_file))
            input_list = [DFTInput(input_file, self.dft_style, flag_symm, None)]

        # for pwmat etot.input: the input is a dict including etot.input file path, kspacing and flag_symm
        elif isinstance(json_input, dict):
            input_file = get_required_parameter("input", json_input)
            if not os.path.isabs(input_file):
                input_file = os.path.join(self.root_dir, input_file)
            if not os.path.exists(input_file):
                raise Exception("Error! The input file {} does not exist!".format(input_file))
            flag_symm = get_parameter("flag_symm", json_input, flag_symm)
            kspacing = get_parameter("kspacing", json_input, None)
            input_list = [DFTInput(input_file, self.dft_style, flag_symm, kspacing)]
        
        # for vasp incar: the input is a str list of multi incar files
        elif isinstance(json_input, list) and isinstance(json_input[0], str):
            for input_file in json_input:
                if not os.path.isabs(input_file):
                    input_file = os.path.join(self.root_dir, input_file)
                if not os.path.exists(input_file):
                    raise Exception("Error! The input file {} does not exist!".format(input_file))
                input_list.append(DFTInput(input_file, self.dft_style, flag_symm, None))

        # for pwmat etot.input: the input is a dict list: for each dict including etot.input file path, kspacing and flag_symm
        elif isinstance(json_input, list) and isinstance(json_input[0], dict):
            for json_detail in json_input:
                input_file = get_required_parameter("input", json_detail)
                if not os.path.isabs(input_file):
                    input_file = os.path.join(self.root_dir, input_file)
                if not os.path.exists(input_file):
                    raise Exception("Error! The etot.input file {} does not exist!".format(input_file))
                flag_symm = get_parameter("flag_symm", json_detail, flag_symm)
                kspacing = get_parameter("kspacing", json_detail, None)
                input_list.append(DFTInput(input_file, self.dft_style, flag_symm, kspacing))
                
        else:
            raise Exception("the dft input file cat not recognized!")

        return input_list

class DFTInput(object):

    def __init__(self, input_file:str, dft_style:str, flag_symm:int, kspacing:int) -> None:
        # super().__init__(input_file=input_file, dft_style=dft_style)
        self.input_file = input_file
        self.dft_style = dft_style
        self.kspacing = kspacing
        self.flag_symm = flag_symm
        self.use_dftb = False
        self.use_skf = False
        self.use_gaussion = False
        # check etot input file
        if self.dft_style == DFT_STYLE.pwmat:
            key_values, etot_lines = read_and_check_etot_input(self.input_file)

            if "MP_N123" in key_values and self.kspacing is not None:
                error_info = "ERROR! The 'kspacing' in DFT/input/{} dict and 'MP_N123' in {} file cannot coexist.\n".format(os.path.basename(self.input_file), os.path.basename(self.input_file))
                error_info += "If 'MP_N123' is not indicated in DFT/input/{}, the 'kspacing' param will be used to generate the 'MP_N123' parameter\n".format(os.path.basename(self.input_file))
                raise Exception(error_info)
            elif "MP_N123" not in key_values and self.kspacing is None:
                self.kspacing  = 0.5

            if "USE_DFTB" in key_values.keys() \
                and key_values["USE_DFTB"] is not None \
                    and key_values["USE_DFTB"] == "T":
                self.use_dftb = True
                if key_values["DFTB_DETAIL"].replace(",", " ").split()[0] != "3": # not chardb
                    self.use_skf = True
            
            if "USE_GAUSSIAN" in key_values.keys() and key_values["USE_GAUSSIAN"]is not None and key_values["USE_GAUSSIAN"] == "T":
                self.use_gaussion

    def get_input_content(self):
        if self.dft_style == DFT_STYLE.pwmat:
            return read_and_check_etot_input(self.input_file)
        elif self.dft_style == DFT_STYLE.vasp:
            with open(self.input_file, "r") as fp:
                lines = fp.readlines()
            return lines
        else:
            raise Exception("the dft style {} not realized!".format(self.dft_style))    
    


import os
import signal
import subprocess
from subprocess import Popen, PIPE
import time
from pwact.active_learning.slurm.slurm import scancle_byjobid
def kill_process(pid:str, job_id:str):
    if job_id is None:
        try:
            # 发送终止信号 (SIGTERM) 给指定的进程
            os.kill(int(pid), signal.SIGTERM)
            print(f"process {pid} has been terminated.")
        except ProcessLookupError:
            print(f"process {pid} non-existent.")
        except PermissionError:
            print(f"No permission to terminate the process {pid}.")
        except Exception as e:
            print(f"Error terminating process {pid}: {e}")
    else:
        scancle_byjobid(job_id)

def get_pid():
    pid = os.getpid()
    job_id = os.getenv('SLURM_JOB_ID')
    return "pid {} job {}".format(pid, job_id) if job_id is not None else "pid {}".format(pid)

